# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Any, Literal
import re
import logging

from dorsal.common.exceptions import ValidationError

logger = logging.getLogger(__name__)

HashFunctionId = Literal["BLAKE3", "SHA-256", "TLSH", "QUICK"]

RX_HEX_64 = re.compile(r"^[0-9a-fA-F]{64}$")
RX_TLSH = re.compile(r"^[tT]{1}1[0-9a-fA-F]{70}$")

RX_LOWERCASE_HEX_64 = re.compile(r"^[0-9a-f]{64}$")
RX_LOWERCASE_HEX_32 = re.compile(r"^[0-9a-f]{32}$")
RX_LOWERCASE_HEX_24 = re.compile(r"^[0-9a-f]{24}$")

RX_MAPPING_HASH_FUNCTION_STRING = {
    "BLAKE3": RX_LOWERCASE_HEX_64,
    "SHA-256": RX_LOWERCASE_HEX_64,
    "TLSH": RX_TLSH,
    "QUICK": RX_LOWERCASE_HEX_64,
}
SUPPORTED_HASH_FUNCTIONS = set(RX_MAPPING_HASH_FUNCTION_STRING.keys())


class HashFunction(Enum):
    SHA256 = "SHA-256"
    BLAKE3 = "BLAKE3"
    QUICK = "QUICK"
    TLSH = "TLSH"


RX_PREFIXED_HASH = re.compile(r"^(?P<prefix>sha-?256|blake3|quick|tlsh):(?P<value>.+)$", re.IGNORECASE)

PREFIX_MAP = {
    "sha-256": HashFunction.SHA256,
    "sha256": HashFunction.SHA256,
    "blake3": HashFunction.BLAKE3,
    "quick": HashFunction.QUICK,
    "tlsh": HashFunction.TLSH,
}


def parse_validate_hash(hash_string: str) -> tuple[str, str] | tuple[None, None]:
    """Returns the normalized (lower-cased) hash string, with hash function identifier.

    Supports: SHA-256, BLAKE3, TLSH and QUICK.

    """
    logger.debug("hash string: %s", hash_string)
    if not isinstance(hash_string, str) or not hash_string:
        logger.debug("error: %s", type(hash_string))
        return None, None

    length = len(hash_string)

    if length == 64:
        return (hash_string.lower(), HashFunction.SHA256.value) if RX_HEX_64.match(hash_string) else (None, None)

    if length == 72 and hash_string.lower().startswith("t"):
        return (hash_string.lower(), HashFunction.TLSH.value) if RX_TLSH.match(hash_string) else (None, None)

    match_ = RX_PREFIXED_HASH.match(hash_string)
    logger.debug("match: %s", match_)
    if not match_:
        return None, None

    groups = match_.groupdict()
    prefix = groups["prefix"].lower()
    value = groups["value"]
    hash_type = PREFIX_MAP.get(prefix)

    logger.debug(
        "groups: %s, prefix: %s, value: %s, hash_type: %s",
        groups,
        prefix,
        value,
        hash_type,
    )

    if not hash_type:
        return None, None

    if hash_type == HashFunction.TLSH:
        if not RX_TLSH.match(value):
            return None, None
    else:
        if not RX_HEX_64.match(value):
            return None, None

    return value.lower(), hash_type.value


class HashStringValidator:
    def is_supported_hash_function(self, candidate: str) -> bool:
        if candidate in RX_MAPPING_HASH_FUNCTION_STRING:
            return True
        return False

    def is_valid(self, candidate_string: str | Any, hash_function: str) -> bool:
        if not isinstance(candidate_string, str):
            return False
        hash_function = hash_function.upper()
        if not RX_MAPPING_HASH_FUNCTION_STRING.get(hash_function.upper()):
            raise ValidationError(f"Unknown/unsupported hash function: {hash_function}")
        return bool(RX_MAPPING_HASH_FUNCTION_STRING[hash_function].match(candidate_string.lower()))

    def validate(self, candidate_string: str, hash_function: str) -> None:
        if self.is_valid(candidate_string=candidate_string, hash_function=hash_function):
            return None
        raise ValidationError(f"Invalid {hash_function} string: {candidate_string}")

    def get_valid_hash(self, candidate_string: str, hash_function: str) -> str | None:
        if self.is_valid(candidate_string=candidate_string, hash_function=hash_function):
            if hash_function == "TLSH":
                return candidate_string.upper()
            return candidate_string.lower()
        return None


hash_string_validator = HashStringValidator()



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import logging
from datetime import datetime, timezone, timedelta
from pydantic import BaseModel, Field
from dorsal.file.utils.dates import PDFDatetime, ensure_aware_datetimes


class DemoModel(BaseModel):
    name: str
    dt: datetime | None = None
    nested: dict[str, datetime] = Field(default_factory=dict)
    tags: list[datetime] = Field(default_factory=list)


@pytest.fixture
def parser():
    return PDFDatetime()


def test_parse_pdf_standard_format(parser):
    # D:YYYYMMDDHHmmSSZ
    assert parser.parse("D:20230101120000Z") == datetime(2023, 1, 1, 12, 0, 0, tzinfo=timezone.utc)

    # Without 'D:' prefix
    assert parser.parse("20230101120000Z") == datetime(2023, 1, 1, 12, 0, 0, tzinfo=timezone.utc)

    # With separators (rare but possible in loose parsing or dateutil fallback)
    # The regex in dates.py is strict about digits, so standard format usually expects compact.
    assert parser.parse("D:20231225") == datetime(2023, 12, 25)


def test_parse_pdf_timezones(parser):
    # UTC-5
    dt = parser.parse("D:20230101120000-05'00'")
    assert dt.hour == 12
    # Check offset
    assert dt.tzinfo is not None
    # -05:00 is 5 hours behind UTC
    assert dt.utcoffset() == timedelta(hours=-5)

    # UTC+1
    dt = parser.parse("D:20230101120000+01'00'")
    assert dt.utcoffset() == timedelta(hours=1)


def test_parse_heroic_recovery(parser):
    # Valid date prefix followed by garbage
    # "20200101" is valid, followed by nonsense numbers
    dt = parser.parse("D:20200101999999")
    assert dt == datetime(2020, 1, 1)

    # Without D: prefix
    dt = parser.parse("20200101garbage")
    assert dt == datetime(2020, 1, 1)


def test_parse_invalid_inputs(parser):
    assert parser.parse(None) is None
    assert parser.parse("") is None
    assert parser.parse("   ") is None
    assert parser.parse(12345) is None  # Not a string

    # Placeholder dates
    assert parser.parse("0001-01-01") is None
    assert parser.parse("None") is None

    # Impossible dates
    assert parser.parse("20231301") is None  # Month 13
    assert parser.parse("20230001") is None  # Month 00


def test_normalize_naive_to_utc():
    naive = datetime(2023, 1, 1, 12, 0)
    result = ensure_aware_datetimes(naive, strategy="utc")

    assert result.tzinfo == timezone.utc
    assert result.year == 2023
    assert result.hour == 12


def test_normalize_list_recursion():
    naive1 = datetime(2023, 1, 1)
    naive2 = datetime(2023, 1, 2)
    data = [naive1, "string", naive2]

    result = ensure_aware_datetimes(data, strategy="utc")

    assert result[0].tzinfo == timezone.utc
    assert result[1] == "string"
    assert result[2].tzinfo == timezone.utc


def test_normalize_dict_recursion():
    naive = datetime(2023, 1, 1)
    data = {"event": naive, "meta": {"created": naive}}

    result = ensure_aware_datetimes(data, strategy="utc")

    assert result["event"].tzinfo == timezone.utc
    assert result["meta"]["created"].tzinfo == timezone.utc


def test_normalize_pydantic_model():
    naive = datetime(2023, 1, 1)
    model = DemoModel(name="test", dt=naive, nested={"key": naive}, tags=[naive])

    result = ensure_aware_datetimes(model, strategy="utc")

    assert result.dt.tzinfo == timezone.utc
    assert result.nested["key"].tzinfo == timezone.utc
    assert result.tags[0].tzinfo == timezone.utc
    # Ensure original is untouched (default in_place=False)
    assert model.dt.tzinfo is None


def test_normalize_in_place_mutation():
    naive = datetime(2023, 1, 1)
    data = {"d": naive}

    # Run in-place
    result = ensure_aware_datetimes(data, strategy="utc", in_place=True)

    assert result["d"].tzinfo == timezone.utc
    assert data["d"].tzinfo == timezone.utc  # Original mutated
    assert result is data  # Reference identity check


def test_normalize_time_soup_warning(caplog):
    """Test that mixing naive and aware timestamps triggers a warning."""
    naive = datetime(2023, 1, 1)
    aware = datetime(2023, 1, 1, tzinfo=timezone(timedelta(hours=1)))

    data = [naive, aware]

    with caplog.at_level(logging.WARNING):
        ensure_aware_datetimes(data, strategy="utc")

    assert "Timezone Consistency Warning" in caplog.text
    assert "Naive dates were coerced" in caplog.text


def test_normalize_multiple_timezones_warning(caplog):
    """Test that having multiple different aware timezones triggers a warning."""
    tz1 = timezone(timedelta(hours=1))
    tz2 = timezone(timedelta(hours=5))

    data = [datetime(2023, 1, 1, tzinfo=tz1), datetime(2023, 1, 1, tzinfo=tz2)]

    with caplog.at_level(logging.WARNING):
        ensure_aware_datetimes(data)

    assert "timestamps from multiple different timezones" in caplog.text


def test_normalize_already_correct():
    """Test that data which is already uniform and aware is processed silently."""
    utc_date = datetime(2023, 1, 1, tzinfo=timezone.utc)
    data = {"a": utc_date, "b": utc_date}

    # Should not warn
    result = ensure_aware_datetimes(data, strategy="utc")
    assert result["a"] == utc_date
    assert result["b"] == utc_date



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import sys
from typing import Any, NoReturn, Sequence

from rich.console import Console
from rich.panel import Panel
from rich.text import Text
from dorsal.common.exceptions import AuthError

import typer


EXIT_CODE_SUCCESS = 0
EXIT_CODE_ERROR = 1

_console_instance: Console | None = None
_error_console_instance: Console | None = None


def get_rich_console() -> Console:
    """Returns a single, shared Console instance, creating it if necessary."""
    global _console_instance
    if _console_instance is None:
        _console_instance = Console()
    return _console_instance


def get_error_console() -> Console:
    """Returns a shared Console instance bound to stderr for diagnostic output."""
    global _error_console_instance
    if _error_console_instance is None:
        _error_console_instance = Console(stderr=True)
    return _error_console_instance


def exit_cli(code: int = EXIT_CODE_SUCCESS, message: str | None = None) -> NoReturn:
    """Comprehensible and testable wrapper for exiting a CLI command.

    Args:
        code: The exit code to use. Defaults to 0 (success).
        message: An optional message to print to stderr before exiting.
                 If the code is > 0, the message will be prefixed with "Error: ".
    """
    if message:
        if code > 0:
            typer.secho(f"Error: {message}", fg=typer.colors.RED, err=True)
        else:
            typer.secho(message, err=True)

    raise typer.Exit(code=code)


def determine_use_cache_value(use_cache: bool, skip_cache: bool) -> bool:
    from dorsal.file.cache.config import get_cache_enabled

    use_cache_choice = None
    if use_cache:
        use_cache_choice = True
    elif skip_cache:
        use_cache_choice = False

    use_cache_value = get_cache_enabled(use_cache=use_cache_choice)

    return use_cache_value


def handle_error(palette: dict, message: str, json_output: bool):
    console = get_rich_console()

    if json_output:
        console.print(json.dumps({"error": True, "detail": message}, indent=2, ensure_ascii=False))
    else:
        panel = Panel(
            Text(message, justify="left"),
            title=f"[{palette.get('panel_title_error', 'bold red')}]Error[/]",
            border_style=palette.get("panel_border_error", "red"),
            expand=False,
            padding=(1, 2),
        )
        console.print(panel)
    exit_cli(code=EXIT_CODE_ERROR)


def handle_auth_error(err: AuthError, console: Console, palette: dict[str, str]) -> None:
    """Handler for AuthError."""
    if "--json" in sys.argv:
        error_payload = {
            "success": False,
            "error": "Authentication Required",
            "detail": "You are not currently logged in.",
            "original_message": str(err),
            "fix": "Run 'dorsal auth login' or set the DORSAL_API_KEY environment variable.",
        }
        print(json.dumps(error_payload, indent=2))
        return

    message = Text.assemble(
        ("You are not currently logged in.\n\n", palette["warning"]),
        ("To authenticate, you can either:\n", "default"),
        ("  1. Run ", "default"),
        ("dorsal auth login\n", f"bold {palette['primary_value']}"),
        ("  2. Set the ", "default"),
        ("DORSAL_API_KEY", f"bold {palette['primary_value']}"),
        (" environment variable.", "default"),
    )

    console.print(
        Panel(
            message,
            expand=False,
            title=f"[{palette['panel_title_info']}]Authentication Required[/]",
            border_style=palette["panel_border_info"],
        )
    )


def handle_offline_error(e: Exception, console: Console, palette: dict):
    """
    Centralized handler for DorsalOfflineError.
    """
    if "--json" in sys.argv:
        error_payload = {
            "success": False,
            "error": "Offline Mode Active",
            "detail": "Communication with DorsalHub is blocked because offline mode is enabled.",
            "original_message": str(e),
            "fix": "Unset the 'DORSAL_OFFLINE' environment variable.",
        }
        print(json.dumps(error_payload, indent=2))
        return

    message = Text.assemble(
        ("Offline Mode is currently active.\n\n", palette["warning"]),
        ("DorsalHub API Access is blocked.\n\n", "default"),
        ("To restore access, unset the", "default"),
        (" DORSAL_OFFLINE", f"bold {palette['primary_value']}"),
        (" environment variable.", "default"),
    )

    console.print(
        Panel(
            message,
            expand=False,
            title=f"[{palette['panel_title_warning']}]Dorsal API Access Blocked[/]",
            border_style=palette["panel_border_info"],
        )
    )


def parse_cli_options(options: Sequence[str] | None, palette: dict) -> dict[str, Any]:
    """Parses a sequence of 'key=value' strings into a dictionary."""
    if not options:
        return {}

    result: dict[str, Any] = {}
    error_console = get_error_console()

    for item in options:
        if "=" not in item:
            error_console.print(
                f"[{palette.get('panel_title_warning')}]Warning:[/] Skipping malformed option '{item}'."
                "Must be in 'key=value' format."
            )
            continue

        key, value = item.split("=", 1)
        key = key.strip()
        value = value.strip()

        val_lower = value.lower()
        if val_lower in ("true", "yes"):
            result[key] = True
        elif val_lower in ("false", "no"):
            result[key] = False
        elif val_lower in ("null", "none"):
            result[key] = None
        else:
            if (value.startswith("{") and value.endswith("}")) or (value.startswith("[") and value.endswith("]")):
                try:
                    result[key] = json.loads(value)
                    continue
                except json.JSONDecodeError:
                    pass

            try:
                if "." in value:
                    result[key] = float(value)
                else:
                    result[key] = int(value)
            except ValueError:
                result[key] = value

    return result


class DummyContext:
    """A no-op context manager to suppress spinners when outputting raw JSON."""

    def __enter__(self):
        return None

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typer
import pathlib
import logging
import datetime
from typing import Annotated, Literal, Optional, cast

logger = logging.getLogger(__name__)


def export_cache_cmd(
    ctx: typer.Context,
    output: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            "--output",
            "-o",
            help="Path to save the exported cache file. Defaults to the '.dorsal/export' directory.",
            writable=True,
            resolve_path=True,
        ),
    ] = None,
    format: Annotated[
        Optional[str],
        typer.Option(
            "--format",
            help="Output format ('json' or 'json.gz'). If not provided, it's inferred from the output file extension.",
            case_sensitive=False,
        ),
    ] = None,
    no_records: Annotated[
        bool,
        typer.Option(
            "--no-records",
            help="Exclude full metadata records from the export to save space.",
        ),
    ] = False,
):
    """
    Exports the full contents of the local cache to a file.
    """
    from dorsal.common.cli import get_rich_console, exit_cli, EXIT_CODE_ERROR
    from dorsal.common import constants
    from dorsal.file.utils.cache import export_cache

    console = get_rich_console()

    palette = ctx.obj["palette"]
    include_records = not no_records

    output_path = output
    if not output_path:
        try:
            constants.CLI_EXPORTS_DIR.mkdir(parents=True, exist_ok=True)
            timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
            output_path = constants.CLI_EXPORTS_DIR / f"cache-export-{timestamp}.json.gz"
        except Exception as e:
            logger.error(f"Failed to create automatic export directory: {e}")
            return exit_cli(
                code=EXIT_CODE_ERROR,
                message=f"Could not create export directory at {constants.CLI_EXPORTS_DIR}.",
            )

    if not format:
        suffix = output_path.suffix.lower()
        if suffix == ".gz" and output_path.stem.endswith(".json"):
            format = "json.gz"
        elif suffix == ".json":
            format = "json"
        else:
            if output:
                console.print(
                    f"[{palette['info']}]Info:[/] No valid format specified or inferred. Defaulting to 'json.gz'."
                )
                output_path = output.with_suffix(".json.gz")
                console.print(f"[{palette['info']}]Info:[/] Output will be saved to: {output_path}")
            format = "json.gz"

    if format not in ["json", "json.gz"]:
        return exit_cli(
            code=EXIT_CODE_ERROR,
            message=f"Invalid format '{format}'. Must be one of 'json' or 'json.gz'.",
        )

    literal_format = cast(Literal["json", "json.gz"], format)
    try:
        with console.status(f"Exporting cache to '{output_path.name}'..."):
            count = export_cache(
                output_path=output_path,
                format=literal_format,
                include_records=include_records,
            )

        console.print(f"[{palette['success']}]✅ Successfully exported {count:,} records to '{output_path}'[/]")

    except (IOError, ValueError) as e:
        exit_cli(code=EXIT_CODE_ERROR, message=f"Export failed: {e}")
    except Exception as e:
        logger.exception("An unexpected error occurred during cache export.")
        exit_cli(code=EXIT_CODE_ERROR, message=f"An unexpected error occurred: {e}")



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import pathlib
import re
from typing import Annotated, Any, Literal, Type, TYPE_CHECKING, Union, cast
import inspect

from pydantic import BaseModel, Field, field_serializer, field_validator

from dorsal.common import constants
from dorsal.common.exceptions import AnnotationImportError
from dorsal.common.model import AnnotationModelSource, AnnotationModel, is_pydantic_model_class
from dorsal.common.validators import (
    CallableImportPath,
    DatasetID,
    import_callable,
    JsonSchemaValidator,
    get_json_schema_validator,
)
from dorsal.file.annotation_models.mediainfo.config import MEDIAINFO_MEDIA_TYPES
from dorsal.file.utils.size import parse_filesize
from dorsal.file.validators.base import MediaTypePartString

logger = logging.getLogger(__name__)

MAX_DEPENDENCY_ARRAY_ITEMS = 100


class RunModelResult(BaseModel):
    """
    The standardized result object returned by `ModelRunner` execution steps.

    This object encapsulates the output of a single Annotation Model, including
    its generated data, source identity, execution timing, and any errors encountered.
    """

    name: str = Field(description="The display name of the model (usually the class name).")
    source: AnnotationModelSource = Field(
        description="Structured metadata identifying the model source (ID, version, variant)."
    )
    record: dict[str, Any] | None = Field(
        description="The generated annotation record (dict). None if the model failed, was skipped, or produced no output."
    )
    schema_id: DatasetID | None = Field(
        description="The validation schema/model ID against which this record was validated."
    )
    schema_version: str | None = Field(
        default=None, description="The version of the schema/model against which this record was validated."
    )
    time_taken: int | float | None = Field(
        default=None, description="Execution time in seconds. Populated only if debug mode is active."
    )
    error: str | None = Field(
        default=None,
        description="A descriptive error message if the model failed, crashed, or if a dependency was not met.",
    )


class DependencyConfig(BaseModel):
    """
    - `type` (str): The primary identifier of the dependency.
    - `checker` (CallableImportPath) - Defines the path to the dependency check function
                           - The dependency check function returns a boolean indicating whether the dependency was met
                           - The dependency check function always takes, as input, the list of prior model outputs (which is always at least the base file model result, at index 0)
    - `silent` (bool): - When set to `False`, raises a `DependencyNotMetError` exception whenever the dependency is not met
                       - When set to `True`, no exception is raised in the case of a dependency not being met.
    """

    type: str
    checker: CallableImportPath
    silent: bool = True


class MediaTypeDependencyConfig(DependencyConfig):
    """This dependency configures which Media Types to execute a model for.

    You can define the match rule for Media Type using any combination of `pattern`, `include`, or `exclude`.

    - `silent` (bool = True): - by default, when not met, the MediaTypeDependencyConfig does not raise.
                              - Set to `False` if you want it to raise an exception
    - `pattern` (str or re.Pattern): match the media type using a regular expression. If the media type matches, the model executes
    - `include` - If the Media Type is in this sequence, the model executes
    - `exclude` - Exclusion rule: if the media type is in this sequence (even if it matches via `pattern` or `include`) it is blocked

    """

    type: Literal["media_type"] = "media_type"
    checker: CallableImportPath = CallableImportPath("dorsal.file.configs.model_runner", "check_media_type_dependency")
    silent: bool = True
    pattern: str | re.Pattern | None = None
    include: set[MediaTypePartString] | None = Field(default=None, max_length=MAX_DEPENDENCY_ARRAY_ITEMS)
    exclude: set[MediaTypePartString] | None = Field(default=None, max_length=MAX_DEPENDENCY_ARRAY_ITEMS)

    @field_validator("include", "exclude", mode="before")
    @classmethod
    def normalize_media_types(cls, v: set[MediaTypePartString] | None) -> set[MediaTypePartString] | None:
        if v is None:
            return None
        return {x.lower() for x in v}

    @field_serializer("include", "exclude")
    def serialize_sets_to_list(self, v: set[str] | None, _info) -> list[str] | None:
        return list(v) if v is not None else None


class FileExtensionDependencyConfig(DependencyConfig):
    """This dependency configures which file extensions to execute a model for."""

    type: Literal["extension"] = "extension"
    checker: CallableImportPath = CallableImportPath("dorsal.file.configs.model_runner", "check_extension_dependency")
    silent: bool = True
    extensions: set[str] = Field(max_length=MAX_DEPENDENCY_ARRAY_ITEMS)

    @field_validator("extensions", mode="before")
    @classmethod
    def normalize_extensions(cls, v: set[str]) -> set[str]:
        if not v:
            return set()
        return {f".{ext.lstrip('.').lower()}" for ext in v}

    @field_serializer("extensions")
    def serialize_extensions_to_list(self, v: set[str], _info) -> list[str]:
        return list(v)


class FileSizeDependencyConfig(DependencyConfig):
    """This dependency configures a model to run based on file size."""

    type: Literal["file_size"] = "file_size"
    checker: CallableImportPath = CallableImportPath("dorsal.file.configs.model_runner", "check_size_dependency")
    silent: bool = True
    min_size: int | None = None
    max_size: int | None = None

    @field_validator("min_size", "max_size", mode="before")
    @classmethod
    def parse_size_strings(cls, v) -> int | None:
        if isinstance(v, str):
            return parse_filesize(v)
        return v


class FilenameDependencyConfig(DependencyConfig):
    """This dependency configures a model to run based on the file's name."""

    type: Literal["file_name"] = "file_name"
    checker: CallableImportPath = CallableImportPath("dorsal.file.configs.model_runner", "check_name_dependency")
    silent: bool = True
    pattern: str | re.Pattern


DependencyType = Union[
    MediaTypeDependencyConfig, FileExtensionDependencyConfig, FileSizeDependencyConfig, FilenameDependencyConfig
]

ModelRunnerDependencyConfig = Annotated[
    DependencyType,
    Field(discriminator="type"),
]


def check_media_type_dependency(model_results: "list[RunModelResult]", config: "MediaTypeDependencyConfig") -> bool:
    """
    Check whether the media type is within the scope of the annotation model.

    - Performs exact string matches or regex matches on the full and partial media type.
    - Checks both *full* and *partial* media type in the following order:
        1. If in `exclude`, return False
        2. If in `include` or matching `pattern`, return True
        3. If `include` or `pattern` were provided, then it failed. Return False
        4. If neither `include` nor `pattern` were provided, then it passed. Return True
    """
    base_record = model_results[0].record

    if not base_record:
        logger.debug("No base record found for media type dependency check")
        return False

    media_type_full = base_record.get("media_type")
    if not media_type_full:
        logger.debug("Base record has no 'media_type' field")
        return False

    media_type_head, _ = media_type_full.split("/")

    if config.exclude:
        if media_type_full in config.exclude or media_type_head in config.exclude:
            logger.debug(f"Media type {media_type_full} is explicitly excluded.")
            return False

    has_inclusion_rules = config.include or config.pattern

    if has_inclusion_rules:
        if config.include:
            if media_type_full in config.include or media_type_head in config.include:
                logger.debug(f"Media type {media_type_full} matched 'include' list.")
                return True

        if config.pattern:
            if isinstance(config.pattern, str):
                rx = re.compile(config.pattern)
            else:
                rx = config.pattern

            m_full = rx.match(media_type_full)
            if m_full:
                logger.debug(f"Media type {media_type_full} matched 'pattern' (full).")
                return True

            m_head = rx.match(media_type_head)
            if m_head:
                logger.debug(f"Media type {media_type_full} matched 'pattern' (head).")
                return True

        logger.debug(f"Media type {media_type_full} did not match any 'include' or 'pattern' rules.")
        return False

    logger.debug(f"Media type {media_type_full} passed (no inclusion rules specified).")
    return True


def check_extension_dependency(model_results: list[RunModelResult], config: FileExtensionDependencyConfig) -> bool:
    """Check whether the extension is within the scope of the annotation model."""
    base_record = model_results[0].record

    if not base_record:
        logger.debug("No base record found")
        return False

    extension: str | None = base_record["extension"]
    if not extension:
        return False

    if extension.lower() in config.extensions:
        return True

    return False


def check_size_dependency(model_results: list[RunModelResult], config: FileSizeDependencyConfig) -> bool:
    """Check whether the file size is within the scope of the annotation model."""
    base_record = model_results[0].record

    if not base_record:
        logger.debug("No base record found for file size dependency check")
        return False

    file_size = base_record.get("size")
    if not isinstance(file_size, int):
        logger.debug("Base record has no valid 'size' field")
        return False

    if config.min_size is not None and file_size < config.min_size:
        logger.debug(f"File size {file_size} is less than min_size {config.min_size}.")
        return False

    if config.max_size is not None and file_size > config.max_size:
        logger.debug(f"File size {file_size} is greater than max_size {config.max_size}.")
        return False

    logger.debug(f"File size {file_size} passed dependency check.")
    return True


def check_name_dependency(model_results: list[RunModelResult], config: FilenameDependencyConfig) -> bool:
    """Check whether the file's name matches the regex pattern."""
    base_record = model_results[0].record

    if not base_record:
        logger.debug("No base record found for filename dependency check")
        return False

    filename = base_record.get("name")
    if not filename:
        logger.debug("Base record has no 'name' field")
        return False

    if isinstance(config.pattern, str):
        rx = re.compile(config.pattern)
    else:
        rx = config.pattern

    if rx.search(filename):
        logger.debug(f"Filename '{filename}' matched pattern '{config.pattern}'.")
        return True

    logger.debug(f"Filename '{filename}' did not match pattern '{config.pattern}'.")
    return False


class ModelRunnerPipelineStep(BaseModel):
    """Single step in the ModelRunner execution pipeline.

    - annotation_model: Two-part path to an Annotation Model.
    - dependencies: Rules to trigger execution.
    - validation_model: Path to validation logic.
    - schema_id: Unique dataset ID.
    - options: Runtime options for the model.
    - ignore_linter_errors: Skip strict linting if True.
    - deactivated: (Optional) If True, this step is skipped. Defaults to False.
    - package_name: The installed package name (used when installing a model via registry)
    """

    annotation_model: CallableImportPath
    dependencies: list[ModelRunnerDependencyConfig] | None = None
    validation_model: CallableImportPath | dict | None = None
    schema_id: DatasetID
    schema_version: str | None = None
    options: dict[str, Any] | None = None
    ignore_linter_errors: bool = False
    deactivated: bool = False
    package_name: str | None = None


BASE_ANNOTATION_MODEL = {
    "annotation_model": ["dorsal.file.annotation_models", "FileCoreAnnotationModel"],
    "validation_model": [
        "dorsal.file.validators.base",
        "FileCoreValidationModelStrict",
    ],
    "schema_id": constants.FILE_BASE_ANNOTATION_SCHEMA,
    "options": {"calculate_similarity_hash": True},
}


def resolve_pipeline_step_models(
    pipeline_step_obj: "ModelRunnerPipelineStep",
) -> tuple[Type[AnnotationModel], Type[BaseModel] | JsonSchemaValidator | None]:
    """
    Resolves the callable model and validator classes from a pipeline step config.
    """
    try:
        annotator_callable = import_callable(import_path=pipeline_step_obj.annotation_model)
        if not (inspect.isclass(annotator_callable) and issubclass(annotator_callable, AnnotationModel)):
            raise TypeError(f"Imported callable '{annotator_callable.__name__}' is not a subclass of AnnotationModel.")
        annotator_class = cast(Type[AnnotationModel], annotator_callable)

        validator: Type[BaseModel] | JsonSchemaValidator | None = None
        if pipeline_step_obj.validation_model:
            if isinstance(pipeline_step_obj.validation_model, dict):
                validator = get_json_schema_validator(schema=pipeline_step_obj.validation_model, strict=True)
            else:
                validator_callable = import_callable(import_path=pipeline_step_obj.validation_model)
                if is_pydantic_model_class(validator_callable):
                    validator = cast(Type[BaseModel], validator_callable)
                elif isinstance(validator_callable, JsonSchemaValidator):
                    validator = validator_callable
                else:
                    raise TypeError("Imported validator is not a supported type.")

        return annotator_class, validator

    except (ImportError, AttributeError, TypeError) as err:
        msg = (
            "Failed to import model/validator from config: "
            f"{pipeline_step_obj.annotation_model.module}.{pipeline_step_obj.annotation_model.name}"
        )
        logger.exception("AnnotationImportError: %s.", msg)
        raise AnnotationImportError(msg) from err



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility module for normalizing, extracting, and validating
language codes, names, and locales.

This module uses 'langcodes' as its primary engine and supports
a custom 'language_code_mapping.json' for overrides and
a 'supported_locales.json' for whitelisting.

Requires: pip install langcodes[data]
"""

import json
import logging
import os
from typing import Dict, Optional, Set, cast
from functools import lru_cache

try:
    from langcodes import Language, tag_is_valid, find as find_lang, LanguageTagError, DEFAULT_LANGUAGE

    _LANGCODES_AVAILABLE = True
except ImportError:
    _LANGCODES_AVAILABLE = False

    class Language:  # type: ignore[no-redef]
        pass

    class LanguageTagError(Exception):  # type: ignore[no-redef]
        pass

    DEFAULT_LANGUAGE = "en"

logger = logging.getLogger(__name__)


_BASE_DIR = os.path.dirname(__file__)
_MAPPING_FILE = os.path.join(_BASE_DIR, "language_mapping.json")


@lru_cache()
def _get_language_map() -> Dict[str, Dict[str, str | None]]:
    """
    Loads the custom language override map from the JSON file.
    This map is used as a "first pass" before consulting langcodes.
    """
    try:
        with open(_MAPPING_FILE, "r", encoding="utf-8") as f:
            return json.load(f)
    except FileNotFoundError:
        logger.info(f"Custom language map not found at {_MAPPING_FILE}. Relying on 'langcodes' library only.")
    except json.JSONDecodeError:
        logger.error(f"Critical: Failed to decode language mapping file {_MAPPING_FILE}")
    except Exception as e:
        logger.error(f"Critical: An unexpected error occurred loading {_MAPPING_FILE}: {e}")
    return {}


@lru_cache()
def get_language_set() -> set[str]:
    """
    Get the set of unique, non-null language *names*
    from the *custom override map*.
    """
    language_map = _get_language_map()
    return {cast(str, v.get("name")) for v in language_map.values() if v and v.get("name")}


@lru_cache()
def get_alpha_3_set() -> set[str]:
    """
    Get the set of unique, non-null language *alpha-3 codes*
    from the *custom override map*.
    """
    language_map = _get_language_map()
    return {cast(str, v.get("alpha_3")) for v in language_map.values() if v and v.get("alpha_3")}


@lru_cache()
def _get_lang_obj(lang_str: str | None) -> Optional[Language]:
    """Get a langcodes.Language object from any input string (code, name, or alias)."""
    if not _LANGCODES_AVAILABLE:
        logger.error("The 'langcodes[data]' library is not installed. Language normalization will not work.")
        return None

    if not lang_str or not isinstance(lang_str, str):
        return None

    clean_str = lang_str.strip()

    language_map = _get_language_map()
    if data := language_map.get(clean_str.lower()):
        code_to_parse = data.get("alpha_3") or data.get("name")
        if code_to_parse:
            try:
                return Language.get(code_to_parse)
            except LanguageTagError:
                pass

    code_str = clean_str.replace("_", "-")
    try:
        return Language.get(code_str)
    except LanguageTagError:
        pass

    try:
        return find_lang(clean_str)
    except (LookupError, NotImplementedError):
        logger.debug(f"Could not find or parse language: '{lang_str}'")
        return None


def normalize_language_name(lang_str: str | None) -> str | None:
    """
    Cleans a messy language string and converts it to its standard
    English language name (e.g., "English", "Chinese (Simplified)").
    """
    lang_obj = _get_lang_obj(lang_str)
    if not lang_obj:
        return None

    try:
        if lang_obj.to_alpha3() == "und":
            return None

        return lang_obj.language_name(DEFAULT_LANGUAGE)
    except Exception:
        return None


def normalize_language_alpha3(lang_str: str | None) -> str | None:
    """
    Cleans a messy language string and converts it to its standard
    ISO 639-3 alpha-3 code (e.g., "eng", "fra", "zho").

    Relies on 'langcodes' library.
    """
    lang_obj = _get_lang_obj(lang_str)
    if not lang_obj:
        return None

    try:
        code = lang_obj.to_alpha3()
        if code == "und":
            return None
        return code
    except LookupError:
        logger.debug(f"Language '{lang_str}' has no alpha-3 code.")
        return None


@lru_cache()
def extract_locale_code(lang_str: str | None) -> str | None:
    """
    Cleans and extracts a BCP-47-like locale code from a messy string.
    This function *identifies* and *cleans* potential locale codes.
    It does not *validate* them against a list of *supported* locales.

    e.g., " en-GB " -> "en-GB"
          "en_US"   -> "en-US"
          "es-419"  -> "es-419"
          "English" -> None (this is a name, not a valid tag)
    """
    if not _LANGCODES_AVAILABLE:
        logger.error("The 'langcodes' library is not installed. Locale extraction will not work.")
        return None

    if not lang_str or not isinstance(lang_str, str):
        return None

    clean_str = lang_str.strip().replace("_", "-")

    if tag_is_valid(clean_str):
        try:
            if Language.get(clean_str).to_alpha3() == "und":
                return None
        except Exception:
            pass

        return clean_str

    logger.debug(f"Could not extract a valid locale code from: '{lang_str}'")
    return None



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import pathlib
import subprocess
import shutil

import typer
from rich.console import Console
from rich.panel import Panel
from rich.text import Text
import json
import os
from typing import Annotated, Optional

logger = logging.getLogger(__name__)

app = typer.Typer(name="auth", help="Manage DorsalHub API Key authentication.", no_args_is_help=True)


def _display_user_info(user_info: dict, title: str, palette: dict, console: Console):
    name = user_info.get("name", "N/A")
    email = user_info.get("email", "N/A")
    account_status = user_info.get("account_status", "N/A")
    user_id = user_info.get("user_id", "N/A")
    info_text = Text.assemble(
        ("User ID:        ", palette["key"]),
        (str(user_id), palette["primary_value"]),
        "\n",
        ("Name:           ", palette["key"]),
        (name, palette["primary_value"]),
        "\n",
        ("Email:          ", palette["key"]),
        (email, palette["primary_value"]),
        "\n",
        ("Account Status: ", palette["key"]),
        (account_status, palette["primary_value"]),
    )
    panel = Panel(
        info_text,
        title=f"[{palette['panel_title_success']}]{title}[/]",
        border_style=palette["panel_border_success"],
        expand=False,
        padding=(1, 2),
    )
    console.print(panel)


@app.command()
def login(
    ctx: typer.Context,
    apikey: Annotated[
        Optional[str],
        typer.Option(
            "--apikey",
            help="Your DorsalHub API key. If not provided, you will be prompted.",
        ),
    ] = None,
    is_project: bool = typer.Option(
        False, "--project", help="Save the API key to the project-level configuration file (e.g., dorsal.toml)."
    ),
):
    """Log in to DorsalHub by providing an API key."""
    from dorsal.session import clear_shared_dorsal_client
    from dorsal.common.auth import write_auth_config
    from dorsal.common.config import find_project_config_path, get_global_config_path, load_config
    from dorsal.common.exceptions import AuthError, DorsalOfflineError, NetworkError
    from dorsal.common.cli import EXIT_CODE_ERROR, exit_cli, get_rich_console
    from dorsal.common import constants

    console = get_rich_console()

    if not apikey:
        console.print("🔑 Please enter your DorsalHub API key.")
        console.print("[dim]You can generate a key from your account settings page.[/dim]")
        apikey = typer.prompt("API Key", hide_input=True)

    if not apikey:
        return exit_cli(code=EXIT_CODE_ERROR, message="API key cannot be empty.")

    try:
        from dorsal.client import DorsalClient

        console.print(":key: Verifying key with DorsalHub...")
        temp_client = DorsalClient(api_key=apikey)
        user_info = temp_client.verify_credentials()

        email = user_info.get("email")
        if is_project:
            scope = "project"
            logger.debug("Writing to the project-level config toml.")
        else:
            scope = "global"
            logger.debug("Writing to the global config at %s", get_global_config_path())

        write_auth_config(api_key=apikey, email=email, scope=scope)

        clear_shared_dorsal_client()

        console.print()
        palette = ctx.obj["palette"]
        _display_user_info(user_info, title="✅ Login Successful", palette=palette, console=console)

        if is_project:
            project_config_path = find_project_config_path()
            if project_config_path:
                config_file_path = project_config_path
            else:
                config_file_path = pathlib.Path.cwd() / "dorsal.toml"
        else:
            config_file_path = get_global_config_path()

        console.print(
            f"\nYour API key has been saved to the {scope} config: [{palette['primary_value']}]{config_file_path}[/]"
        )

        if is_project:
            config_filename = config_file_path.name
            console.print()
            security_warning_text = Text.assemble(
                ("The file ", "default"),
                (f"'{config_filename}'", f"bold {palette['primary_value']}"),
                (" now contains your API key.\n\n", "default"),
                ("This file ", palette["warning"]),
                ("must not", palette["error"]),
                (" be committed to Git.\n", palette["warning"]),
                ("To automatically add this file to your .gitignore, please run:", "default"),
                ("\n  dorsal auth gitignore", f"bold {palette['primary_value']}"),
            )
            console.print(
                Panel(
                    security_warning_text,
                    title=f"[{palette['panel_title_warning']}]🔒 Action Required[/]",
                    border_style=palette["panel_border_warning"],
                    expand=False,
                    padding=(1, 2),
                )
            )

        exit_cli()
    except DorsalOfflineError:
        raise
    except AuthError:
        exit_cli(code=EXIT_CODE_ERROR, message="The provided API key is invalid or expired.")
    except NetworkError as err:
        exit_cli(code=EXIT_CODE_ERROR, message=f"Could not connect to DorsalHub. {err}")
    except OSError as err:
        exit_cli(code=EXIT_CODE_ERROR, message=f"Could not save API Key. {err}")


@app.command()
def logout(
    ctx: typer.Context,
):
    """
    Log out by removing the API key from the local configuration.
    """
    from dorsal.session import clear_shared_dorsal_client
    from dorsal.common.auth import get_api_key_details, remove_api_key, APIKeySource
    from dorsal.common.cli import EXIT_CODE_ERROR, exit_cli, get_rich_console

    console = get_rich_console()
    palette = ctx.obj["palette"]

    try:
        details = get_api_key_details()
        active_source = details["source"]

        if active_source == APIKeySource.ENV:
            warning_message = Text.assemble(
                ("You are currently authenticated via an environment variable.\n\n", palette["warning"]),
                ("The ", "default"),
                ("DORSAL_API_KEY", f"bold {palette['primary_value']}"),
                (" is set. This command only removes keys from configuration files.\n\n", "default"),
                ("To complete the logout process for this session, please unset this environment variable.", "default"),
            )
            console.print(
                Panel(
                    warning_message,
                    title=f"[{palette['panel_title_warning']}]Environment Variable Active[/]",
                    border_style=palette["panel_border_warning"],
                )
            )
            console.print("No configuration files were changed.")
            return exit_cli()

        if active_source == APIKeySource.NONE:
            console.print("You are not currently logged in.")
            return exit_cli()

        key_was_removed = remove_api_key(scope=active_source)
        clear_shared_dorsal_client()

        if key_was_removed:
            console.print(f"[{palette['success']}]✅ You have been successfully logged out.[/]")
        else:
            console.print("You are not currently logged in (or the key could not be removed).")
        return exit_cli()

    except OSError as err:
        console.print(f"[{palette['error']}]Error:[/] Could not modify config file. {err}")
        return exit_cli(code=EXIT_CODE_ERROR, message=f"Could not modify config file. {err}")


@app.command()
def gitignore(ctx: typer.Context):
    """
    Check if the project config file is in .gitignore, and add it if not.
    """
    from dorsal.common.cli import EXIT_CODE_ERROR, exit_cli, get_rich_console
    from dorsal.common.config import get_project_level_config
    from dorsal.common import constants

    console = get_rich_console()
    palette = ctx.obj["palette"]
    logger.debug("--- Running 'dorsal auth gitignore' ---")

    if not shutil.which("git"):
        logger.warning("'git' command not found.")
        console.print(f"[{palette['warning']}]'git' command not found. Cannot check or update .gitignore.[/]")
        return exit_cli(code=EXIT_CODE_ERROR)

    current_cwd = pathlib.Path.cwd()
    logger.debug(f"Checking for git repo in CWD: {current_cwd}")
    try:
        root_result_args = ["git", "rev-parse", "--show-toplevel"]
        root_result = subprocess.run(
            root_result_args, capture_output=True, text=True, check=True, cwd=current_cwd, encoding="utf-8"
        )
        repo_root = pathlib.Path(root_result.stdout.strip())
        logger.debug(f"Git root found at: {repo_root}")

    except (subprocess.CalledProcessError, FileNotFoundError) as e:
        logger.debug(f"Not a git repo (or 'git rev-parse' failed): {e}")
        console.print(f"[{palette['info']}]This directory does not appear to be a Git repository. Nothing to do.[/]")
        return exit_cli()

    project_config_data, project_config_path = get_project_level_config()
    logger.debug(f"'get_project_level_config' returned path: {project_config_path}")

    if not project_config_path:
        logger.debug("No project-level config file found.")
        console.print(f"[{palette['info']}]No project-level config file found. Nothing to do.[/]")
        return exit_cli()

    if not project_config_data:
        logger.debug(f"Project config file found ('{project_config_path.name}'), but it is empty or invalid.")
        console.print(
            f"[{palette['info']}]Project config file '{project_config_path.name}' is empty or invalid. Nothing to do.[/]"
        )
        return exit_cli()

    config_filename = project_config_path.name
    logger.debug(f"Config filename basename: {config_filename}")

    auth_section_present = project_config_data.get(constants.CONFIG_SECTION_AUTH) is not None

    if not auth_section_present:
        logger.debug("Config file found, but it contains no 'auth' section.")
        console.print(
            f"[{palette['success']}]✅ Your '{config_filename}' file contains no 'auth' section.[/]", highlight=False
        )
        console.print("It is safe to commit and share with your team. No changes made to .gitignore.")
        exit_cli()

    logger.debug("Config file contains an 'auth' section. Proceeding with .gitignore check.")

    gitignore_path = repo_root / ".gitignore"
    logger.debug(f"Target .gitignore path: {gitignore_path}")

    found_in_gitignore = False
    if gitignore_path.exists():
        logger.debug(f"'{gitignore_path}' exists. Reading contents.")
        with open(gitignore_path, "r", encoding="utf-8") as f:
            lines = f.readlines()

        for i, line in enumerate(lines):
            stripped_line = line.strip()
            if stripped_line == config_filename or stripped_line == f"/{config_filename}":
                logger.debug(f"Found matching entry on line {i + 1}: '{stripped_line}'")
                found_in_gitignore = True
                break
    else:
        logger.debug(f"'{gitignore_path}' does not exist.")

    if found_in_gitignore:
        logger.debug("Entry already exists. No action taken.")
        console.print(f"[{palette['success']}]✅ Entry for '{config_filename}' already exists in .gitignore.[/]")

        note_text = Text.assemble(
            ("Note: If Git is still tracking this file, run the following command to stop tracking it:", "default"),
            (f"\n  git rm --cached {config_filename}", palette["primary_value"]),
        )
        console.print(
            Panel(
                note_text,
                title=f"[{palette['panel_title_info']}]File Still Tracked?[/]",
                border_style=palette["panel_border_info"],
                expand=False,
                padding=(1, 2),
            )
        )
        exit_cli()

    logger.debug("File is not in .gitignore. Asking for user consent.")
    console.print(f"The config file '[{palette['primary_value']}]{config_filename}[/]' is not in your .gitignore.")

    try:
        if not typer.confirm("Do you want to add it to the project's root .gitignore file?"):
            logger.debug("User consent denied (N).")
            console.print("Operation cancelled. Please add the file to .gitignore manually.")
            exit_cli()

        logger.debug("User consent granted (y).")
    except typer.Abort:
        logger.debug("User aborted consent (Ctrl+C).")
        console.print("\nOperation cancelled.")
        exit_cli()

    logger.debug("Proceeding to write to .gitignore.")
    try:
        entry_to_add = f"\n# Dorsal config file (contains secrets)\n/{config_filename}\n"
        logger.debug(f"Entry to add: {entry_to_add.strip()}")

        with open(gitignore_path, "a", encoding="utf-8") as f:
            f.write(entry_to_add)

        logger.debug("Successfully wrote to .gitignore.")
        console.print(f"[{palette['success']}]✅ Successfully added '/{config_filename}' to {gitignore_path}.[/]")
        exit_cli()

    except (OSError, subprocess.CalledProcessError) as e:
        logger.error(f"Failed during .gitignore write operation: {e}", exc_info=True)
        console.print(f"[{palette['error']}]Error:[/] Could not write to .gitignore: {e}")
        console.print("Please add the file to .gitignore manually.")
        exit_cli(code=EXIT_CODE_ERROR)


@app.command()
def whoami(
    ctx: typer.Context,
    json_output: Annotated[
        bool,
        typer.Option(
            "--json",
            help="Output the user info as a raw JSON object to stdout for scripting.",
        ),
    ] = False,
):
    """
    Check the currently authenticated user and session status.
    """
    from dorsal.session import get_shared_dorsal_client
    from dorsal.common.cli import EXIT_CODE_ERROR, exit_cli, get_rich_console
    from dorsal.common.exceptions import AuthError, NetworkError

    console = get_rich_console()

    try:
        client = get_shared_dorsal_client()

        if not json_output:
            console.print("Verifying session with DorsalHub...")

        user_info = client.verify_credentials()

        if json_output:
            console.print(json.dumps(user_info, indent=2, default=str, ensure_ascii=False))
        else:
            palette = ctx.obj["palette"]
            console.print()
            _display_user_info(
                user_info,
                title="👤 Authenticated User",
                palette=palette,
                console=console,
            )
        exit_cli()

    except NetworkError as err:
        exit_cli(code=EXIT_CODE_ERROR, message=f"Could not connect to DorsalHub. {err}")



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import datetime
import json
import logging
import pathlib
from typing import Annotated, TYPE_CHECKING, Optional

from rich.panel import Panel
from rich.table import Table
from rich.markup import escape
import typer

from dorsal.common import constants

if TYPE_CHECKING:
    from dorsal.api.file import _DirectoryInfoResult

logger = logging.getLogger(__name__)


def info_directory(
    ctx: typer.Context,
    path: Annotated[
        pathlib.Path,
        typer.Argument(
            exists=True,
            file_okay=False,
            dir_okay=True,
            readable=True,
            help="The directory path to analyze.",
        ),
    ],
    recursive: Annotated[
        bool,
        typer.Option(
            "--recursive/--no-recursive",
            "-r/-R",
            help="Scan subdirectories recursively.",
            rich_help_panel="Scan Options",
        ),
    ] = False,
    media_type: Annotated[
        bool,
        typer.Option(
            "--media-type",
            "-m",
            help="Include Media Type summary table. Reduces scan speed.",
            rich_help_panel="Scan Options",
        ),
    ] = False,
    json_output: Annotated[
        bool,
        typer.Option(
            "--json",
            help="Output a raw JSON object to stdout for scripting.",
            rich_help_panel="Output Options",
        ),
    ] = False,
    save: Annotated[
        bool,
        typer.Option(
            "-s",
            "--save",
            help="Save the JSON report to the default directory or --output path.",
            rich_help_panel="Output Options",
        ),
    ] = False,
    output_path: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            "-o",
            "--output",
            help="Custom path to save the JSON report (e.g., 'info.json').",
            dir_okay=True,
            file_okay=True,
            writable=True,
            resolve_path=True,
            rich_help_panel="Output Options",
        ),
    ] = None,
):
    """
    Displays summary of files in a directory.
    """
    from dorsal.api.file import get_directory_info
    from dorsal.common.cli import EXIT_CODE_ERROR, get_rich_console, exit_cli
    from dorsal.file.utils.size import human_filesize

    console = get_rich_console()
    palette = ctx.obj["palette"]

    if output_path and not save:
        if str(output_path).lower().endswith(".json"):
            save = True
        else:
            if not json_output:
                console.print(
                    f"⚠️ [yellow]Warning:[/] --output path '{output_path}' was specified with an unknown extension."
                    f" Please use -s (for .json) or specify a .json file.",
                    style=palette.get("warning", "yellow"),
                )

    progress_console = None if json_output else console
    try:
        dir_info = get_directory_info(
            dir_path=str(path),
            recursive=recursive,
            media_type=media_type,
            progress_console=progress_console,
            palette=palette,
        )

        successfully_processed = dir_info["overall"]["total_size"] > 0 or dir_info["overall"]["total_files"] > 0
        if not dir_info or not successfully_processed:
            if not json_output:
                console.print(f"[{palette['warning']}]⚠️ No files found or accessible in '{escape(str(path))}'.[/]")
            exit_cli()

        if json_output:
            console.print(json.dumps(dir_info, indent=2, default=str, ensure_ascii=False))
            exit_cli()

        overall = dir_info["overall"]
        duration_val = overall["time_taken_seconds"]
        duration_str = f"{duration_val:.2f} seconds" if duration_val >= 0.01 else "< 0.01 seconds"
        summary_table = Table.grid(expand=False)
        summary_table.add_column(justify="right", style=palette["key"], width=24)
        summary_table.add_column(justify="left", style=palette["primary_value"])
        summary_table.add_row("Total File Count:", f"{overall['total_files']:,}")
        summary_table.add_row("Total Directories:", f"{overall['total_dirs']:,}")
        summary_table.add_row("Hidden Files:", f"{overall['hidden_files']:,}")
        summary_table.add_row("Total Size:", human_filesize(overall["total_size"]))
        summary_table.add_row("Scan Duration:", duration_str)
        summary_table.add_row()
        summary_table.add_row("Average File Size:", human_filesize(overall["avg_size"]))
        largest_path = escape(overall["largest_file"]["path"]) if overall["largest_file"]["path"] else "N/A"
        smallest_path = escape(overall["smallest_file"]["path"]) if overall["smallest_file"]["path"] else "N/A"
        summary_table.add_row(
            "Largest File:",
            f"{largest_path} ({human_filesize(overall['largest_file']['size'])})",
        )
        summary_table.add_row(
            "Smallest File:",
            f"{smallest_path} ({human_filesize(overall['smallest_file']['size'])})",
        )
        summary_table.add_row()

        def format_date_row(data: dict):
            path = escape(data["path"]) if data["path"] else "N/A"
            date_str = "N/A"
            if data["date"]:
                date_obj = datetime.datetime.fromisoformat(data["date"])
                date_str = date_obj.strftime("%Y-%m-%d %H:%M:%S")
            return f"{date_str} ({path})"

        summary_table.add_row("Newest Modified File:", format_date_row(overall["newest_mod_file"]))
        summary_table.add_row("Oldest Modified File:", format_date_row(overall["oldest_mod_file"]))
        summary_table.add_row("Oldest Creation Date:", format_date_row(overall["oldest_creation_file"]))
        permissions = overall.get("permissions", {})
        if permissions:
            summary_table.add_row()
            summary_table.add_row("Executable Files:", f"{permissions.get('executable', 0):,}")
            summary_table.add_row("Read-Only Files:", f"{permissions.get('read_only', 0):,}")

        console.print(f"📊 Summary of [{palette['primary_value']}]{escape(str(path))}[/]")
        console.print(
            Panel(
                summary_table,
                title=f"[{palette['panel_title']}]Directory Summary[/]",
                expand=False,
                border_style=palette["panel_border"],
            )
        )

        if media_type and dir_info["by_type"]:
            by_type_table = Table(
                title="Media Type Breakdown",
                show_header=True,
                header_style=palette["table_header"],
                expand=False,
            )
            by_type_table.add_column("Media Type", style=palette["primary_value"], ratio=55)
            by_type_table.add_column("Count", justify="right", ratio=15)
            by_type_table.add_column("Total Size", justify="right", ratio=15)
            by_type_table.add_column("% of Total", justify="right", ratio=15)
            for item in dir_info["by_type"]:
                by_type_table.add_row(
                    item["media_type"],
                    f"{item['count']:,}",
                    human_filesize(item["total_size"]),
                    f"{item['percentage']:.2f}%",
                )
            console.print(by_type_table)

        if save:
            _save_json_report(
                dir_info=dir_info,
                source_path=path,
                output_path=output_path,
                palette=palette,
                json_to_stdout=json_output,
            )

    except (FileNotFoundError, NotADirectoryError) as e:
        exit_cli(code=EXIT_CODE_ERROR, message=str(e))
    except typer.Exit:
        raise
    except Exception as err:
        logger.exception("CLI 'info' command failed.")
        exit_cli(code=EXIT_CODE_ERROR, message=str(err))


def _get_final_path(source_path: pathlib.Path, output_path: Optional[pathlib.Path], suffix: str) -> pathlib.Path:
    """Helper to determine the final save path for a report."""

    if output_path:
        if output_path.is_dir():
            dir_name = source_path.name
            return output_path / f"stats-{dir_name}{suffix}"
        else:
            return output_path

    constants.CLI_STATS_REPORTS_DIR.mkdir(parents=True, exist_ok=True)
    timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    dir_name = source_path.name.replace(" ", "_")
    return constants.CLI_STATS_REPORTS_DIR / f"stats-{dir_name}-{timestamp}{suffix}"


def _save_json_report(
    dir_info: _DirectoryInfoResult,
    source_path: pathlib.Path,
    output_path: Optional[pathlib.Path],
    palette: dict,
    json_to_stdout: bool,
):
    """Saves the dir_info dictionary to a JSON file."""
    from dorsal.common.cli import EXIT_CODE_ERROR, get_rich_console, exit_cli

    console = get_rich_console()

    final_path = _get_final_path(source_path, output_path, ".json")

    try:
        final_path.parent.mkdir(parents=True, exist_ok=True)
        with open(final_path, "w", encoding="utf-8") as f:
            json.dump(dir_info, f, default=str, indent=2, ensure_ascii=False)

        if not json_to_stdout:
            console.print(f"✅ JSON report saved to: [{palette.get('primary_value')}]{final_path}[/]")
    except IOError as e:
        logger.exception(f"Failed to write to output file: {final_path}")
        exit_cli(code=EXIT_CODE_ERROR, message=f"Error writing to file: {e}")
    except Exception as e:
        logger.error(f"Failed to save JSON report: {e}")
        console.print(
            f"⚠️ Could not save JSON report to {final_path}. Error: {e}", style=palette.get("warning", "yellow")
        )



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import json
import pathlib
from typer.testing import CliRunner
from unittest.mock import MagicMock, call
from rich.panel import Panel


from dorsal.cli import app


from dorsal.cli.dir_app import duplicates_dir_cmd


runner = CliRunner()

TEST_DATA_DIR = "tests/data"

MOCK_DUPLICATES_RESULT = {
    "path": str(TEST_DATA_DIR),
    "total_sets": 1,
    "hashes_from_cache": 5,
    "duplicate_sets": [
        {
            "hash": "hash_val_1",
            "hash_type": "sha256",
            "count": 2,
            "file_size": "10.0 KB",
            "file_size_bytes": 10240,
            "paths": [f"{TEST_DATA_DIR}/file_a.txt", f"{TEST_DATA_DIR}/file_b.txt"],
        }
    ],
}


@pytest.fixture
def mock_duplicates_cmd(mocker):
    """Mocks backend dependencies for the `dir duplicates` command."""
    import dorsal.api.file

    mocker.patch("dorsal.common.cli.determine_use_cache_value", return_value=True)

    mock_find_dupes = mocker.patch.object(dorsal.api.file, "find_duplicates", return_value=MOCK_DUPLICATES_RESULT)

    mock_save_report = mocker.patch.object(duplicates_dir_cmd, "_save_duplicates_report")

    return {
        "find_duplicates": mock_find_dupes,
        "save_report": mock_save_report,
    }


def test_duplicates_dir_success_panel_output(mock_rich_console, mock_duplicates_cmd):
    """Tests the default `dir duplicates` command, expecting Rich Panel output."""
    result = runner.invoke(app, ["dir", "duplicates", TEST_DATA_DIR])

    assert result.exit_code == 0, f"CLI exited with error: {result.output}"

    printed_items = [call.args[0] for call in mock_rich_console.print.call_args_list]
    assert any(isinstance(item, Panel) for item in printed_items), "No Rich Panel was printed."

    mock_duplicates_cmd["find_duplicates"].assert_called_once()
    mock_duplicates_cmd["save_report"].assert_not_called()


def test_duplicates_dir_no_duplicates(mock_rich_console, mock_duplicates_cmd):
    """Tests the output when no duplicates are found."""
    mock_duplicates_cmd["find_duplicates"].return_value = {}

    result = runner.invoke(app, ["dir", "duplicates", TEST_DATA_DIR])

    assert result.exit_code == 0, f"CLI exited with error: {result.output}"

    all_printed_text = "".join(str(call.args[0]) for call in mock_rich_console.print.call_args_list)
    assert "No duplicate files found" in all_printed_text


def test_duplicates_dir_json_output(mock_rich_console, mock_duplicates_cmd):
    """Tests the --json output flag."""
    result = runner.invoke(app, ["dir", "duplicates", TEST_DATA_DIR, "--json"])

    assert result.exit_code == 0, f"CLI exited with error: {result.output}"

    mock_rich_console.print.assert_called_once()
    json_output_str = mock_rich_console.print.call_args.args[0]
    data = json.loads(json_output_str)

    assert data["total_sets"] == 1
    mock_duplicates_cmd["save_report"].assert_not_called()


def test_duplicates_dir_limit_output(mock_rich_console, mock_duplicates_cmd):
    """Tests that the --limit flag correctly limits the number of printed panels."""
    multi_result = {
        **MOCK_DUPLICATES_RESULT,
        "total_sets": 2,
        "duplicate_sets": [
            MOCK_DUPLICATES_RESULT["duplicate_sets"][0],
            {
                "hash": "hash_val_2",
                "paths": ["c.txt", "d.txt"],
                "file_size": "5.0 KB",
                "hash_type": "sha256",
                "count": 2,
                "file_size_bytes": 5120,
            },
        ],
    }
    mock_duplicates_cmd["find_duplicates"].return_value = multi_result

    result = runner.invoke(app, ["dir", "duplicates", TEST_DATA_DIR, "--limit", "1"])

    assert result.exit_code == 0, f"CLI exited with error: {result.output}"
    printed_panels = [item for item in mock_rich_console.print.call_args_list if isinstance(item.args[0], Panel)]
    assert len(printed_panels) == 1

    all_printed_text = "".join(str(call.args[0]) for call in mock_rich_console.print.call_args_list)
    assert "Displaying the" in all_printed_text and "1" in all_printed_text


def test_duplicates_dir_exception_handling(mock_rich_console, mock_duplicates_cmd):
    """Tests that a generic exception is handled gracefully."""
    mock_duplicates_cmd["find_duplicates"].side_effect = Exception("File system is unreadable")

    result = runner.invoke(app, ["dir", "duplicates", TEST_DATA_DIR])

    assert result.exit_code != 0

    assert "File system is unreadable" in result.output



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

__all__ = [
    "DorsalFile",
    "FileAnnotationStub",
    "LocalFile",
    "ModelRunner",
    "MetadataReader",
    "scan_file",
    "scan_directory",
    "index_file",
    "index_directory",
    "generate_html_file_report",
    "get_blake3_hash",
    "get_quick_hash",
    "get_sha256_hash",
]

if TYPE_CHECKING:
    from dorsal.file.model_runner import ModelRunner
    from dorsal.file.metadata_reader import MetadataReader
    from dorsal.file.dorsal_file import DorsalFile, FileAnnotationStub, LocalFile
    from dorsal.file.utils import get_blake3_hash, get_quick_hash, get_sha256_hash
    from dorsal.api.file import scan_file, scan_directory, index_file, index_directory, generate_html_file_report


def __getattr__(name: str):
    """
    Lazily import heavy file processing classes and API functions to prevent
    'fat init' overhead when simply traversing the dorsal.file namespace.
    """
    if name == "ModelRunner":
        from dorsal.file.model_runner import ModelRunner

        return ModelRunner

    elif name == "MetadataReader":
        from dorsal.file.metadata_reader import MetadataReader

        return MetadataReader

    elif name == "DorsalFile":
        from dorsal.file.dorsal_file import DorsalFile

        return DorsalFile

    elif name == "LocalFile":
        from dorsal.file.dorsal_file import LocalFile

        return LocalFile

    elif name == "FileAnnotationStub":
        from dorsal.file.dorsal_file import FileAnnotationStub

        return FileAnnotationStub

    elif name in {"get_blake3_hash", "get_quick_hash", "get_sha256_hash"}:
        from dorsal.file.utils import get_blake3_hash, get_quick_hash, get_sha256_hash

        return locals()[name]

    elif name in {"scan_file", "scan_directory", "index_file", "index_directory", "generate_html_file_report"}:
        from dorsal.api.file import scan_file, scan_directory, index_file, index_directory, generate_html_file_report

        return locals()[name]

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    return __all__



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import json
from typer.testing import CliRunner
from unittest.mock import MagicMock
from rich.table import Table
from rich.panel import Panel


from dorsal.cli import app
from dorsal.common.exceptions import DorsalClientError, NotFoundError

runner = CliRunner()
COLLECTION_ID = "col_abc123"


mock_file = MagicMock()
mock_file.name = "research_paper.pdf"
mock_file.size = 1048576
mock_file.media_type = "application/pdf"
mock_file.hash = "sha256:mockhash123"

mock_collection = MagicMock()
mock_collection.file_count = 1


mock_pagination = MagicMock()
mock_pagination.page_count = 1


MOCK_API_RESPONSE = MagicMock()
MOCK_API_RESPONSE.collection = mock_collection
MOCK_API_RESPONSE.files = [mock_file]
MOCK_API_RESPONSE.pagination = mock_pagination
MOCK_API_RESPONSE.model_dump_json.return_value = json.dumps(
    {
        "collection": {"id": COLLECTION_ID, "file_count": 1},
        "files": [{"name": "research_paper.pdf"}],
    }
)


@pytest.fixture
def mock_show_collection_cmd(mocker):
    """Mocks dependencies for the `collection show` command."""
    mock_get = mocker.patch("dorsal.api.collection.get_collection", return_value=MOCK_API_RESPONSE)

    mock_view = mocker.patch(
        "dorsal.cli.views.collection.collection_metadata",
        return_value=Panel("Mock Metadata Panel"),
    )

    return {
        "get_collection": mock_get,
        "view": mock_view,
    }


def test_show_collection_success_default(mock_rich_console, mock_show_collection_cmd):
    """Tests the default successful command, expecting a panel and a table."""
    result = runner.invoke(app, ["collection", "show", COLLECTION_ID])

    assert result.exit_code == 0
    mock_show_collection_cmd["get_collection"].assert_called_once_with(
        collection_id=COLLECTION_ID, hydrate=False, page=1, per_page=30, mode="pydantic"
    )
    mock_show_collection_cmd["view"].assert_called_once()

    assert mock_rich_console.print.call_count == 3
    assert isinstance(mock_rich_console.print.call_args_list[1].args[0], Panel)
    assert isinstance(mock_rich_console.print.call_args_list[2].args[0], Table)


def test_show_collection_meta_only(mock_rich_console, mock_show_collection_cmd):
    """Tests that --meta-only shows the panel but not the file table."""
    result = runner.invoke(app, ["collection", "show", COLLECTION_ID, "--meta-only"])

    assert result.exit_code == 0
    assert mock_show_collection_cmd["get_collection"].call_args.kwargs["per_page"] == 0

    assert mock_rich_console.print.call_count == 2
    assert isinstance(mock_rich_console.print.call_args.args[0], Panel)


def test_show_collection_no_files(mock_rich_console, mock_show_collection_cmd):
    """Tests the output when a collection is empty."""
    MOCK_API_RESPONSE.files = []
    MOCK_API_RESPONSE.collection.file_count = 0

    result = runner.invoke(app, ["collection", "show", COLLECTION_ID])

    assert result.exit_code == 0
    assert mock_rich_console.print.call_count == 3
    assert "Collection contains no files" in mock_rich_console.print.call_args.args[0]


def test_show_collection_json_output(mock_rich_console, mock_show_collection_cmd):
    """Tests the --json flag, expecting raw JSON output."""
    result = runner.invoke(app, ["collection", "show", COLLECTION_ID, "--json"])

    assert result.exit_code == 0
    # hydrate should be True for JSON output
    assert mock_show_collection_cmd["get_collection"].call_args.kwargs["hydrate"] is True

    mock_rich_console.print.assert_called_once()
    MOCK_API_RESPONSE.model_dump_json.assert_called_once()
    # The view helper should NOT be called in JSON mode
    mock_show_collection_cmd["view"].assert_not_called()


def test_show_collection_not_found(mock_show_collection_cmd):
    """Tests that a NotFoundError is handled gracefully."""
    mock_show_collection_cmd["get_collection"].side_effect = NotFoundError("test")

    result = runner.invoke(app, ["collection", "show", COLLECTION_ID])

    assert result.exit_code != 0
    assert f"Collection '{COLLECTION_ID}' not found" in result.output


def test_show_collection_api_error(mock_show_collection_cmd):
    """Tests that a generic DorsalClientError is handled gracefully."""
    mock_show_collection_cmd["get_collection"].side_effect = DorsalClientError("Permission Denied")

    result = runner.invoke(app, ["collection", "show", COLLECTION_ID])

    assert result.exit_code != 0
    assert "API Error: Permission Denied" in result.output



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import sqlite3
import zlib
import os
import time
from unittest.mock import patch, MagicMock
from pathlib import Path

from dorsal.file.cache.dorsal_cache import DorsalCache, CachedFileRecord


@pytest.fixture
def temp_cache(tmp_path: Path) -> DorsalCache:
    """
    Provides a clean DorsalCache instance pointed to a unique temporary
    database file for each test. The tmp_path fixture is managed by pytest.
    """
    db_path = tmp_path / "test_cache.db"
    cache = DorsalCache(db_path=db_path)
    cache.connect()
    yield cache
    cache.close()


@pytest.fixture
def mock_file_record_strict() -> MagicMock:
    """Provides a mock FileRecordStrict object for testing insertions."""
    record = MagicMock()
    record.annotations.file_base.record.name = "test.pdf"
    record.annotations.file_base.record.size = 1024
    record.annotations.file_base.record.extension = ".pdf"
    record.annotations.file_base.record.media_type = "application/pdf"
    record.annotations.file_base.record.all_hash_ids = {
        "SHA-256": "a" * 64,
        "QUICK": "b" * 64,
    }
    record.model_dump_json.return_value = '{"hash": "a...'
    return record


# --- Tests ---


def test_initialization_and_schema(temp_cache: DorsalCache):
    """Test that the database table and indexes are created on connection."""
    cursor = temp_cache.conn.cursor()
    cursor.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='cached_files'")
    assert cursor.fetchone() is not None

    cursor.execute("SELECT name FROM sqlite_master WHERE type='index' AND name='idx_hash_sha256'")
    assert cursor.fetchone() is not None


def test_upsert_and_get_record_uncompressed(temp_cache: DorsalCache, mock_file_record_strict):
    """Test inserting and retrieving a record with compression disabled."""
    temp_cache.use_compression = False
    path, mtime = "/fake/file.pdf", 12345.6789

    temp_cache.upsert_record(path=path, modified_time=mtime, record=mock_file_record_strict)
    cached_record = temp_cache.get_record(path=path)

    assert isinstance(cached_record, CachedFileRecord)
    assert cached_record.abspath == path
    assert cached_record.name == "test.pdf"


def test_upsert_and_get_record_compressed(temp_cache: DorsalCache, mock_file_record_strict):
    """Test that records are correctly compressed and decompressed."""
    temp_cache.use_compression = True
    path, mtime = "/fake/file.pdf", 12345.6789

    temp_cache.upsert_record(path=path, modified_time=mtime, record=mock_file_record_strict)
    cached_record = temp_cache.get_record(path=path)

    assert cached_record is not None
    assert cached_record.record_json == '{"hash": "a...'

    cursor = temp_cache.conn.cursor()
    cursor.execute("SELECT is_compressed FROM cached_files WHERE abspath = ?", (path,))
    assert cursor.fetchone()["is_compressed"] == 1


def test_get_record_not_found(temp_cache: DorsalCache):
    """Test that get_record returns None for a path that doesn't exist."""
    assert temp_cache.get_record(path="/does/not/exist.txt") is None


def test_upsert_and_get_hash_success(temp_cache: DorsalCache, fs):
    """Test inserting a single hash and retrieving it successfully."""
    file_path = "/fake/file.txt"
    fs.create_file(file_path)
    current_mtime = os.path.getmtime(file_path)

    temp_cache.upsert_hash(
        path=file_path,
        modified_time=current_mtime,
        hash_function="SHA-256",
        hash_value="a" * 64,
    )

    retrieved_hash = temp_cache.get_hash(path=file_path, hash_function="SHA-256")
    assert retrieved_hash == "a" * 64


def test_get_hash_stale_record(temp_cache: DorsalCache, fs):
    """Test that get_hash returns None if the file's modification time has changed."""
    file_path = "/fake/file.txt"
    fs.create_file(file_path)
    old_mtime = os.path.getmtime(file_path)

    temp_cache.upsert_hash(
        path=file_path,
        modified_time=old_mtime,
        hash_function="SHA-256",
        hash_value="a" * 64,
    )

    time.sleep(0.01)  # Ensure timestamp will be different
    new_mtime = time.time()
    os.utime(file_path, (new_mtime, new_mtime))

    retrieved_hash = temp_cache.get_hash(path=file_path, hash_function="SHA-256")
    assert retrieved_hash is None


def test_clear(temp_cache: DorsalCache):
    """Test that the clear method deletes the database file."""
    db_path = temp_cache.db_path
    assert db_path.exists()

    temp_cache.clear()

    assert not db_path.exists()
    assert temp_cache.conn is None


def test_summary(temp_cache: DorsalCache, mock_file_record_strict):
    """Test that the cache summary provides accurate counts."""
    # 1. Insert one full record
    temp_cache.upsert_record(path="/fake/full.pdf", modified_time=123.45, record=mock_file_record_strict)

    # 2. Insert one hash-only record
    temp_cache.upsert_hash(
        path="/fake/hash_only.txt",
        modified_time=123.45,
        hash_function="SHA-256",
        hash_value="b" * 64,
    )

    summary = temp_cache.summary()

    assert summary["total_records"] == 2
    assert summary["full_records"] == 1
    assert summary["hash_only_records"] == 1
    assert summary["database_size_bytes"] > 0


def test_prune(temp_cache: DorsalCache, fs):
    """Test that prune correctly removes records for deleted or modified files."""
    # 1. Setup the fake filesystem and cache records
    fs.create_file("/fake/current.txt")
    fs.create_file("/fake/stale.txt")
    fs.create_file("/fake/deleted.txt")

    mtime_current = os.path.getmtime("/fake/current.txt")
    mtime_stale = os.path.getmtime("/fake/stale.txt")
    mtime_deleted = os.path.getmtime("/fake/deleted.txt")

    temp_cache.upsert_hash(
        path="/fake/current.txt",
        modified_time=mtime_current,
        hash_function="SHA-256",
        hash_value="a" * 64,
    )
    temp_cache.upsert_hash(
        path="/fake/stale.txt",
        modified_time=mtime_stale,
        hash_function="SHA-256",
        hash_value="b" * 64,
    )
    temp_cache.upsert_hash(
        path="/fake/deleted.txt",
        modified_time=mtime_deleted,
        hash_function="SHA-256",
        hash_value="c" * 64,
    )

    # 2. Alter the filesystem
    os.remove("/fake/deleted.txt")
    time.sleep(0.01)
    # Update the modification time of the stale file
    with open("/fake/stale.txt", "w") as f:
        f.write("new content")

    # 3. Run prune and check the results
    pruned_count, total_records = temp_cache.prune()

    assert pruned_count == 2
    assert total_records == 3

    # 4. Verify the correct records remain in the DB
    assert temp_cache.get_hash(path="/fake/current.txt") is not None
    assert temp_cache.get_hash(path="/fake/stale.txt") is None
    assert temp_cache.get_hash(path="/fake/deleted.txt") is None


def test_close_connection(temp_cache: DorsalCache):
    """Test that the close method closes the database connection."""
    connection = temp_cache.conn
    assert connection is not None

    temp_cache.close()

    assert temp_cache.conn is None
    with pytest.raises(sqlite3.ProgrammingError, match="Cannot operate on a closed database."):
        connection.execute("SELECT 1")


def test_vacuum(temp_cache: DorsalCache):
    """Test that the vacuum method executes the VACUUM command."""
    executed_sql = []

    def trace_callback(statement):
        executed_sql.append(statement)

    # Use the built-in sqlite3 tracer to record executed commands
    temp_cache.conn.set_trace_callback(trace_callback)

    temp_cache.vacuum()

    # Assert that the VACUUM command was present in the executed statements
    assert "VACUUM" in " ".join(executed_sql)


@patch("dorsal.file.cache.dorsal_cache.os.path.getsize")
def test_optimize(mock_getsize, temp_cache: DorsalCache):
    """Test the optimize method to ensure it orchestrates other methods correctly."""
    # Mock the methods that optimize calls
    temp_cache.prune = MagicMock(return_value=(5, 100))  # 5 pruned
    temp_cache._sync_compression = MagicMock(return_value=10)  # 10 rewritten
    temp_cache.vacuum = MagicMock()

    # Mock the file size before and after
    mock_getsize.side_effect = [2048, 1024]  # size_before, size_after

    result = temp_cache.optimize()

    # Assert that all underlying methods were called
    temp_cache.prune.assert_called_once()
    temp_cache._sync_compression.assert_called_once()
    temp_cache.vacuum.assert_called_once()

    # Assert the summary is calculated correctly
    assert result["stale_records_removed"] == 5
    assert result["records_rewritten_for_compression"] == 10
    assert result["size_before_bytes"] == 2048
    assert result["size_after_bytes"] == 1024
    assert result["size_reclaimed_bytes"] == 1024


def test_sync_compression_compresses_records(temp_cache: DorsalCache):
    """Test that _sync_compression correctly compresses uncompressed records."""
    temp_cache.use_compression = True
    uncompressed_data = b'{"test": "data"}'

    # Manually insert an uncompressed record
    cursor = temp_cache.conn.cursor()
    cursor.execute(
        "INSERT INTO cached_files (abspath, modified_time, record, is_compressed) VALUES (?, ?, ?, ?)",
        ("/fake/file.txt", 123.45, uncompressed_data, 0),
    )
    temp_cache.conn.commit()

    rewritten_count = temp_cache._sync_compression()
    assert rewritten_count == 1

    # Verify the record is now compressed
    cursor.execute(
        "SELECT record, is_compressed FROM cached_files WHERE abspath = ?",
        ("/fake/file.txt",),
    )
    row = cursor.fetchone()
    assert row["is_compressed"] == 1
    assert zlib.decompress(row["record"]) == uncompressed_data


def test_sync_compression_decompresses_records(temp_cache: DorsalCache):
    """Test that _sync_compression correctly decompresses compressed records."""
    temp_cache.use_compression = False
    uncompressed_data = b'{"test": "data"}'
    compressed_data = zlib.compress(uncompressed_data)

    # Manually insert a compressed record
    cursor = temp_cache.conn.cursor()
    cursor.execute(
        "INSERT INTO cached_files (abspath, modified_time, record, is_compressed) VALUES (?, ?, ?, ?)",
        ("/fake/file.txt", 123.45, compressed_data, 1),
    )
    temp_cache.conn.commit()

    rewritten_count = temp_cache._sync_compression()
    assert rewritten_count == 1

    # Verify the record is now uncompressed
    cursor.execute(
        "SELECT record, is_compressed FROM cached_files WHERE abspath = ?",
        ("/fake/file.txt",),
    )
    row = cursor.fetchone()
    assert row["is_compressed"] == 0
    assert row["record"] == uncompressed_data



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations
import json
import typer
import pathlib
import datetime
import time
from typing import Annotated, TYPE_CHECKING, Literal, Optional

from pydantic import BaseModel, ValidationError
from rich.panel import Panel
from rich.table import Table
from rich.markup import escape
from rich.text import Text
import logging

from dorsal.common import constants

if TYPE_CHECKING:
    from dorsal.file.dorsal_file import LocalFile
    from dorsal.file.collection.local import LocalFileCollection

logger = logging.getLogger(__name__)


class SortBy(BaseModel):
    value: Literal["name", "size", "type", "date"]


class SortOrder(BaseModel):
    value: Literal["asc", "desc"]


def scan_directory(
    ctx: typer.Context,
    path: Annotated[
        pathlib.Path,
        typer.Argument(
            exists=True,
            file_okay=False,
            dir_okay=True,
            readable=True,
            help="The path to the directory to scan.",
        ),
    ],
    output_path: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            "--output",
            "-o",
            help="Path to save an output file. Extension (.json or .csv) determines format.",
            dir_okay=True,
            file_okay=True,
            writable=True,
            resolve_path=True,
            rich_help_panel="Output Options",
        ),
    ] = None,
    save: Annotated[
        bool,
        typer.Option(
            "-s",
            "--save",
            help="Save the full JSON metadata report to the default scan directory.",
            rich_help_panel="Output Options",
        ),
    ] = False,
    csv: Annotated[
        bool,
        typer.Option(
            "-c",
            "--csv",
            help="Save the summary file table as a CSV report.",
            rich_help_panel="Output Options",
        ),
    ] = False,
    json_output: Annotated[
        bool,
        typer.Option(
            "--json",
            help="Output the full metadata as a raw JSON object to stdout for scripting.",
            rich_help_panel="Output Options",
        ),
    ] = False,
    recursive: Annotated[
        bool,
        typer.Option(
            "--recursive",
            "-r",
            help="Scan subdirectories recursively.",
            rich_help_panel="Scan Options",
        ),
    ] = False,
    limit: Annotated[
        int,
        typer.Option(
            "--limit",
            "-l",
            help="Limit the number of files displayed in the summary table.",
            rich_help_panel="Scan Options",
        ),
    ] = 20,
    sort_by: Annotated[
        str,
        typer.Option(
            case_sensitive=False,
            help="Column to sort by. One of: name, size, type, date.",
            rich_help_panel="Scan Options",
        ),
    ] = "name",
    sort_order: Annotated[
        str,
        typer.Option(
            case_sensitive=False,
            help="Sort order. One of: asc, desc.",
            rich_help_panel="Scan Options",
        ),
    ] = "asc",
    use_cache: Annotated[
        bool,
        typer.Option(
            "--use-cache",
            help="Force the use of the cache, overriding any global setting.",
            rich_help_panel="Cache Options",
        ),
    ] = False,
    skip_cache: Annotated[
        bool,
        typer.Option(
            "--skip-cache",
            help="Bypass the local cache and re-process all files, overriding any global setting.",
            rich_help_panel="Cache Options",
        ),
    ] = False,
    overwrite_cache: Annotated[
        bool,
        typer.Option(
            "--overwrite-cache",
            help="Re-process all files and overwrite the local cache with new data.",
            rich_help_panel="Cache Options",
        ),
    ] = False,
    resolve_links: Annotated[
        bool,
        typer.Option(
            "--follow-links/--no-follow-links",
            help="Follow symlinks to scan target content vs scanning the link itself.",
        ),
    ] = True,
    lazy: Annotated[
        bool,
        typer.Option(
            "--lazy",
            help="Start processing immediately with an indeterminate progress bar (spinner). Useful for massive directories.",
            rich_help_panel="Scan Options",
        ),
    ] = False,
):
    """
    Scans a directory, generates metadata for all files, and displays or saves the results.
    """
    from dorsal.common.cli import (
        EXIT_CODE_ERROR,
        get_rich_console,
        exit_cli,
        determine_use_cache_value,
    )
    from dorsal.file.collection.local import LocalFileCollection

    console = get_rich_console()
    palette: dict[str, str] = ctx.obj["palette"]

    start_time = time.perf_counter()

    try:
        SortBy(value=sort_by)
        SortOrder(value=sort_order)
    except ValidationError as e:
        exit_cli(code=EXIT_CODE_ERROR, message=f"Invalid sorting option provided: {e}")

    if use_cache and skip_cache:
        exit_cli(
            code=EXIT_CODE_ERROR,
            message="Error: --use-cache and --skip-cache flags cannot be used together.",
        )

    if skip_cache and overwrite_cache:
        exit_cli(
            code=EXIT_CODE_ERROR,
            message="Error: --skip-cache and --overwrite-cache flags cannot be used together.",
        )

    if output_path and not (save or csv):
        if str(output_path).lower().endswith(".json"):
            save = True
        elif str(output_path).lower().endswith(".csv"):
            csv = True
        else:
            console.print(
                f"Warning --output path '{output_path}' was specified with an unknown extension."
                f" Please use -s (for .json) or -c (for .csv) to specify the report type.",
                style=palette.get("warning", "yellow"),
            )

    use_cache_value = determine_use_cache_value(use_cache=use_cache, skip_cache=skip_cache)

    try:
        progress_console = None if json_output else console
        collection = LocalFileCollection(
            source=str(path),
            console=progress_console,
            palette=palette,
            recursive=recursive,
            use_cache=use_cache_value,
            overwrite_cache=overwrite_cache,
            follow_symlinks=resolve_links,
            lazy=lazy,
        )
    except Exception as e:
        logger.exception("Failed to initialize FileCollection.")
        exit_cli(
            code=EXIT_CODE_ERROR,
            message=f"An error occurred during file discovery: {e}",
        )

    duration = time.perf_counter() - start_time

    if json_output:
        scan_data = {
            "scan_metadata": {
                "path": str(path),
                "recursive": recursive,
                "duration_seconds": duration,
                "total_files_found": len(collection),
            },
            "results": collection.to_dict(),
        }
        console.print(json.dumps(scan_data, indent=2, default=str))
        exit_cli()

    collection_info = collection.info()
    source_breakdown: list[dict] = collection_info.get("by_source", [])
    files_from_cache = 0
    for source_stat in source_breakdown:
        if source_stat.get("source") == "cache":
            files_from_cache = source_stat.get("count", 0)
            break

    cache_info_str = (
        f" ([{palette.get('success', 'green')}]{files_from_cache} from cache[/])" if files_from_cache > 0 else ""
    )

    console.print(
        f"Found and processed [{palette['success']}]{len(collection)}[/] file(s) in [{palette['primary_value']}]{escape(str(path))}[/]{cache_info_str} in {duration:.3f} seconds."
    )

    if collection.warnings:
        warning_panel = Panel(
            "\n".join(f"- {w}" for w in collection.warnings),
            title=f"[{palette['panel_title_warning']}]Warnings[/]",
            border_style=palette["panel_border_warning"],
            title_align="left",
            expand=False,
        )
        console.print(warning_panel)

    if not collection:
        exit_cli()

    _print_directory_summary_panel(collection_info=collection_info, palette=palette)
    _print_file_details_table(
        collection=collection,
        palette=palette,
        limit=limit,
        sort_by=sort_by,
        sort_order=sort_order,
    )

    if save:
        _save_json_report(
            collection=collection,
            output_path=output_path,
            palette=palette,
        )

    if csv:
        _save_csv_report(collection=collection, output_path=output_path, palette=palette)


def _get_final_path(source_path: pathlib.Path, output_path: Optional[pathlib.Path], suffix: str) -> pathlib.Path:
    """Helper to determine the final save path for a report."""

    if output_path:
        if output_path.is_dir():
            dir_name = source_path.name
            return output_path / f"scan-dir-{dir_name}{suffix}"
        else:
            return output_path

    constants.CLI_SCAN_REPORTS_DIR.mkdir(parents=True, exist_ok=True)
    timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    dir_name = source_path.name.replace(" ", "_")
    return constants.CLI_SCAN_REPORTS_DIR / f"scan-dir-{dir_name}-{timestamp}{suffix}"


def _save_json_report(
    collection: "LocalFileCollection",
    output_path: Optional[pathlib.Path],
    palette: dict,
):
    """Saves the full JSON metadata report."""
    from dorsal.common.cli import EXIT_CODE_ERROR, get_rich_console, exit_cli

    console = get_rich_console()

    source_path = pathlib.Path(collection.source_info.get("path", "unknown-dir"))
    final_path = _get_final_path(source_path, output_path, ".json")

    try:
        final_path.parent.mkdir(parents=True, exist_ok=True)
        collection.to_json(str(final_path), exclude={"embeddings", "text_chunks"})
        console.print(f"✅ JSON report saved to: [{palette.get('primary_value')}]{final_path}[/]")
    except IOError as e:
        logger.exception(f"Failed to write to output file: {final_path}")
        exit_cli(code=EXIT_CODE_ERROR, message=f"Error writing to file: {e}")
    except Exception as e:
        logger.error(f"Failed to save JSON report: {e}")
        console.print(
            f"⚠️ Could not save JSON report to {final_path}. Error: {e}", style=palette.get("warning", "yellow")
        )


def _save_csv_report(
    collection: "LocalFileCollection",
    output_path: Optional[pathlib.Path],
    palette: dict,
):
    """Saves the summary table as a CSV report."""
    from dorsal.common.cli import EXIT_CODE_ERROR, get_rich_console, exit_cli

    console = get_rich_console()

    source_path = pathlib.Path(collection.source_info.get("path", "unknown-dir"))
    final_path = _get_final_path(source_path, output_path, ".csv")

    console.print("\n📄 Saving table as CSV report...")
    try:
        final_path.parent.mkdir(parents=True, exist_ok=True)
        collection.to_csv(
            str(final_path),
        )
        console.print(f"✅ CSV report saved to: [{palette.get('primary_value')}]{final_path}[/]")
    except IOError as e:
        logger.exception(f"Failed to write to output file: {final_path}")
        exit_cli(code=EXIT_CODE_ERROR, message=f"Error writing to file: {e}")
    except Exception as e:
        logger.error(f"Failed to save CSV report: {e}")
        console.print(
            f"⚠️ Could not save CSV report to {final_path}. Error: {e}", style=palette.get("warning", "yellow")
        )


def _print_directory_summary_panel(collection_info: dict, palette: dict):
    """Prints a rich summary panel of the directory's contents."""
    from dorsal.common.cli import get_rich_console
    from dorsal.file.utils.size import human_filesize

    console = get_rich_console()

    overall = collection_info.get("overall", {})
    by_type = collection_info.get("by_type", [])

    total_files = overall.get("total_files", 0)
    total_size = human_filesize(overall.get("total_size", 0))
    media_type_count = len(by_type)

    newest = overall.get("newest_file", {})
    oldest = overall.get("oldest_file", {})

    newest_str = (
        f"{newest['date'].strftime('%Y-%m-%d %H:%M:%S')} ({escape(newest['path'])})" if newest.get("path") else "N/A"
    )
    oldest_str = (
        f"{oldest['date'].strftime('%Y-%m-%d %H:%M:%S')} ({escape(oldest['path'])})" if oldest.get("path") else "N/A"
    )

    summary_text = Text(no_wrap=True)
    summary_text.append("       Total Files: ", style=palette.get("key"))
    summary_text.append(str(total_files), style=palette.get("value"))
    summary_text.append("\n")
    summary_text.append("        Total Size: ", style=palette.get("key"))
    summary_text.append(total_size, style=palette.get("value"))
    summary_text.append("\n")
    summary_text.append("Newest Modified File: ", style=palette.get("key"))
    summary_text.append(newest_str, style=palette.get("value"))
    summary_text.append("\n")
    summary_text.append("Oldest Modified File: ", style=palette.get("key"))
    summary_text.append(oldest_str, style=palette.get("value"))
    summary_text.append("\n\n")
    summary_text.append("       Media Types: ", style=palette.get("key"))
    summary_text.append(str(media_type_count), style=palette.get("value"))

    panel = Panel(
        summary_text,
        title=f"[{palette.get('panel_title', 'bold default')}]Directory Scan Summary[/]",
        border_style=palette.get("panel_border", "blue"),
        title_align="left",
        expand=False,
    )
    console.print(panel)


def _print_file_details_table(
    collection: "LocalFileCollection",
    palette: dict,
    limit: int,
    sort_by: str,
    sort_order: str,
):
    """Prints a rich, detailed table of individual files to the console."""
    from dorsal.common.cli import get_rich_console
    from dorsal.file.utils.size import human_filesize

    console = get_rich_console()

    total_files = len(collection)

    sort_key_map = {
        "name": lambda file: file.name.lower(),
        "size": lambda file: file.size,
        "type": lambda file: file.media_type,
        "date": lambda file: file.date_modified,
    }
    key_func = sort_key_map[sort_by]
    is_reverse = sort_order == "desc"
    sorted_files = sorted(list(collection), key=key_func, reverse=is_reverse)
    files_to_display = sorted_files[:limit]

    table = Table(
        title="File Scan Details",
        show_header=True,
        header_style=palette["table_header"],
        expand=False,
    )

    table.add_column("Filename", style=palette["primary_value"], min_width=30, overflow="ellipsis")
    table.add_column("Size", justify="right", style=palette.get("value"))
    table.add_column("Media Type", style=palette.get("value"))
    table.add_column("Modified Date", style=palette.get("value"))

    for file in files_to_display:
        modified_date_str = file.date_modified.strftime("%Y-%m-%d %H:%M:%S")

        path_obj = pathlib.Path(file._file_path)
        display_name = file.name

        if path_obj.is_symlink():
            try:
                target = path_obj.readlink()
                display_name = f"{escape(path_obj.name)} [dim italic]→ {escape(str(target))}[/]"
            except OSError:
                display_name = f"{escape(path_obj.name)} [dim italic](symlink)[/]"

        table.add_row(
            display_name,
            human_filesize(file.size),
            file.media_type,
            modified_date_str,
        )

    console.print(table)

    if total_files > limit:
        console.print(
            f"[{palette.get('info', 'dim')}]Showing first {limit} of {total_files} files. Use --limit to show more or save the full report.[/]"
        )



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Iterator, List, Union


def get_file_paths(dir_path: str, recursive: bool = False, lazy: bool = False) -> Union[List[str], Iterator[str]]:
    """
    Retrieves file paths from a directory.

    Args:
        dir_path: The directory to scan.
        recursive: If True, scans subdirectories recursively.
        lazy: If True, yields paths, else returns a list after scanning is complete.

    Returns:
        List[str] | Iterator[str]: A list of file paths or an iterator yielding them.
    """

    def _scanner() -> Iterator[str]:
        if recursive:
            for root, _, files in os.walk(dir_path):
                for file in files:
                    yield os.path.join(root, file)
        else:
            try:
                with os.scandir(dir_path) as entries:
                    for entry in entries:
                        if entry.is_file():
                            yield entry.path
            except OSError:
                return

    if lazy:
        return _scanner()

    return list(_scanner())



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
import zipfile
import xml.etree.ElementTree as ET
from typing import Any

from dorsal.file.utils.dates import PDF_DATETIME

logger = logging.getLogger(__name__)

RELS_NS = "{http://schemas.openxmlformats.org/package/2006/relationships}"
CT_NS = "{http://schemas.openxmlformats.org/package/2006/content-types}"
CP_NS = "{http://schemas.openxmlformats.org/officeDocument/2006/custom-properties}"
VT_NS = "{http://schemas.openxmlformats.org/officeDocument/2006/docPropsVTypes}"


BASE_NS_MAP = {
    "cp": "http://schemas.openxmlformats.org/package/2006/metadata/core-properties",
    "dc": "http://purl.org/dc/elements/1.1/",
    "dcterms": "http://purl.org/dc/terms/",
    "ep": "http://schemas.openxmlformats.org/officeDocument/2006/extended-properties",
    "ct": "http://schemas.openxmlformats.org/package/2006/content-types",
    "cpcp": "http://schemas.openxmlformats.org/officeDocument/2006/custom-properties",
    "vt": "http://schemas.openxmlformats.org/officeDocument/2006/docPropsVTypes",
}

CORE_PROPS_TYPE = "http://schemas.openxmlformats.org/package/2006/relationships/metadata/core-properties"
APP_PROPS_TYPE = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/extended-properties"

DEFAULT_CORE_PROPS_PATH = "docProps/core.xml"
DEFAULT_APP_PROPS_PATH = "docProps/app.xml"

KEYWORD_SPLIT_RX = re.compile(r"[;,]")


def _split_keywords(kw_str: str | None) -> list[str]:
    if not kw_str:
        return []
    return [k.strip() for k in KEYWORD_SPLIT_RX.split(kw_str) if k.strip()]


def _find_xml_element(xml_root: ET.Element | None, namespaced_path: str, ns_map: dict[str, str]) -> ET.Element | None:
    if xml_root is None:
        return None

    el = xml_root.find(namespaced_path, ns_map)

    if el is None:
        local_name = namespaced_path.split(":")[-1]
        el = xml_root.find(local_name)

    return el


def _findall_xml_elements(
    xml_root: ET.Element | None, namespaced_path: str, ns_map: dict[str, str]
) -> list[ET.Element]:
    if xml_root is None:
        return []

    elements = xml_root.findall(namespaced_path, ns_map)

    if not elements:
        local_name = namespaced_path.split(":")[-1]
        elements = xml_root.findall(local_name)

    return elements


def _get_xml_text(xml_root: ET.Element | None, namespaced_path: str, ns_map: dict[str, str]) -> str | None:
    el = _find_xml_element(xml_root, namespaced_path, ns_map)
    if el is not None and el.text:
        return el.text.strip()
    return None


def _get_xml_text_as_int(xml_root: ET.Element | None, namespaced_path: str, ns_map: dict[str, str]) -> int | None:
    text = _get_xml_text(xml_root, namespaced_path, ns_map)
    if text and text.isdigit():
        return int(text)
    return None


def _parse_date(date_str: str | None) -> Any:
    if not date_str or not PDF_DATETIME:
        return None
    return PDF_DATETIME.parse(date_str)


def _find_relationship_target(rels_root: ET.Element | None, type_url: str) -> str | None:
    """Parses a .rels file to find the 'Target' for a given 'Type'."""
    if rels_root is None:
        return None

    for el in rels_root.findall(f"{RELS_NS}Relationship"):
        if el.get("Type") == type_url:
            target = el.get("Target")
            if target:
                return target.lstrip("/")
    return None


def _read_xml_from_zip(z: zipfile.ZipFile, path: str) -> ET.Element | None:
    try:
        xml_data = z.read(path)
        return ET.fromstring(xml_data)
    except KeyError:
        logger.debug(f"[v6] XML file not found in zip: '{path}'")
    except ET.ParseError as e:
        logger.warning(f"[v6] Failed to parse XML file '{path}': {e}")
    except Exception as e:
        logger.error(f"[v6] Unexpected error reading '{path}': {e}", exc_info=True)
    return None


def _parse_core_properties(xml_root: ET.Element | None, ns_map: dict[str, str]) -> dict[str, Any]:
    """Extracts metadata from the core.xml element tree."""
    if xml_root is None:
        return {}
    return {
        "author": _get_xml_text(xml_root, "dc:creator", ns_map),
        "last_modified_by": _get_xml_text(xml_root, "cp:lastModifiedBy", ns_map),
        "title": _get_xml_text(xml_root, "dc:title", ns_map),
        "subject": _get_xml_text(xml_root, "dc:subject", ns_map),
        "keywords": _split_keywords(_get_xml_text(xml_root, "cp:keywords", ns_map)),
        "revision": _get_xml_text_as_int(xml_root, "cp:revision", ns_map),
        "creation_date": _parse_date(_get_xml_text(xml_root, "dcterms:created", ns_map)),
        "modified_date": _parse_date(_get_xml_text(xml_root, "dcterms:modified", ns_map)),
    }


def _parse_app_properties(xml_root: ET.Element | None, ns_map: dict[str, str]) -> tuple[dict[str, Any], dict[str, Any]]:
    """Extracts metadata from the app.xml element tree."""
    if xml_root is None:
        return {}, {}

    app_data = {
        "application_name": _get_xml_text(xml_root, "ep:Application", ns_map),
        "application_version": _get_xml_text(xml_root, "ep:AppVersion", ns_map),
        "template": _get_xml_text(xml_root, "ep:Template", ns_map),
    }
    word_specific_data = {
        "page_count": _get_xml_text_as_int(xml_root, "ep:Pages", ns_map),
        "word_count": _get_xml_text_as_int(xml_root, "ep:Words", ns_map),
        "char_count": _get_xml_text_as_int(xml_root, "ep:Characters", ns_map),
        "paragraph_count": _get_xml_text_as_int(xml_root, "ep:Paragraphs", ns_map),
    }
    return app_data, word_specific_data


def _parse_content_types_xml(xml_root: ET.Element | None) -> dict[str, Any]:
    """Extracts structural parts and has_comments from [Content_Types].xml."""
    if xml_root is None:
        return {}
    parts = []
    has_comments = False

    for el in xml_root.findall(f"{CT_NS}Default") + xml_root.findall(f"{CT_NS}Override"):
        part = el.get("ContentType")
        if part:
            parts.append(part)
            if not has_comments and "comments" in part:
                has_comments = True

    return {"structural_parts": sorted(list(set(parts))), "has_comments": has_comments}


def _parse_custom_properties_xml(xml_root: ET.Element | None, ns_map: dict[str, str]) -> dict[str, Any]:
    """Extracts key-value pairs from custom.xml."""
    if xml_root is None:
        return {}
    props = {}

    for el in _findall_xml_elements(xml_root, "cpcp:property", ns_map):
        name = el.get("name")
        if not name:
            continue

        val_el = next(iter(el), None)
        if val_el is not None and val_el.text:
            props[name] = val_el.text.strip()

    return {"custom_properties": props}



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import typer
import pathlib
from typing import Annotated, List, Literal
import json

from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from rich.markup import escape

from dorsal.common.literals import MiB


logger = logging.getLogger(__name__)

HashType = Literal["BLAKE3", "SHA-256", "TLSH", "QUICK"]


def hash_file(
    ctx: typer.Context,
    path: Annotated[
        pathlib.Path,
        typer.Argument(
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="The path to the single file to hash.",
        ),
    ],
    sha256: Annotated[bool, typer.Option("--sha256", help="Display the SHA-256 hash.")] = False,
    blake3: Annotated[bool, typer.Option("--blake3", help="Display the BLAKE3 hash.")] = False,
    tlsh: Annotated[bool, typer.Option("--tlsh", help="Display the TLSH similarity hash.")] = False,
    quick: Annotated[bool, typer.Option("--quick", help="Display the sample-based QuickHash.")] = False,
    use_cache: Annotated[
        bool,
        typer.Option(
            "--use-cache",
            help="Force the use of the cache, overriding any global setting.",
            rich_help_panel="Cache Options",
        ),
    ] = False,
    skip_cache: Annotated[
        bool,
        typer.Option(
            "--skip-cache",
            help="Bypass the local cache and re-calculate hashes, overriding any global setting.",
            rich_help_panel="Cache Options",
        ),
    ] = False,
    json_output: Annotated[bool, typer.Option("--json", help="Output hashes as a raw JSON object.")] = False,
):
    """
    Calculates and displays cryptographic and specialized hashes for a single file.
    """
    from dorsal.common.cli import (
        EXIT_CODE_ERROR,
        get_rich_console,
        determine_use_cache_value,
        exit_cli,
    )
    from dorsal.file.hash_reader import HASH_READER
    from dorsal.file.utils.quick_hasher import QuickHasher

    console = get_rich_console()
    palette = ctx.obj["palette"]

    if use_cache and skip_cache:
        exit_cli(
            code=EXIT_CODE_ERROR,
            message="Error: --use-cache and --skip-cache flags cannot be used together.",
        )

    use_cache_value = determine_use_cache_value(use_cache=use_cache, skip_cache=skip_cache)

    hashes_to_get: List[HashType] = []
    show_all = not any([sha256, blake3, tlsh, quick])

    if show_all or sha256:
        hashes_to_get.append("SHA-256")
    if show_all or blake3:
        hashes_to_get.append("BLAKE3")
    if show_all or tlsh:
        hashes_to_get.append("TLSH")
    if show_all or quick:
        hashes_to_get.append("QUICK")

    try:
        file_hashes = HASH_READER.get(
            file_path=str(path),
            hashes=list(set(hashes_to_get)),
            skip_cache=not use_cache_value,
        )
    except Exception as err:
        logger.exception(f"CLI 'hash' command failed for path {path}.")
        return exit_cli(code=EXIT_CODE_ERROR, message=str(err))

    if json_output:
        console.print(json.dumps(file_hashes, indent=2, ensure_ascii=False, default=str))
        return exit_cli()

    if len(hashes_to_get) == 1:
        hash_type = hashes_to_get[0]
        hash_value = file_hashes.get(hash_type)

        if hash_value is not None:
            console.print(hash_value)
        elif hash_type == "QUICK":
            min_size_mb = QuickHasher.min_permitted_filesize // MiB
            console.print(
                f"[{palette['info']}]Info:[/] QuickHash not generated. File size is below the {min_size_mb}MiB minimum."
            )
        elif hash_type == "TLSH":
            console.print(
                f"[{palette['info']}]Info:[/] TLSH hash not generated. To enable, please install the 'py-tlsh' Python package."
            )
        return exit_cli()

    if file_hashes:
        hash_grid = Table.grid(padding=(0, 2))
        hash_grid.add_column(justify="right", style=palette["key"])
        hash_grid.add_column(justify="left")

        for hash_function, hash_val in file_hashes.items():
            if hash_function == "TLSH" and hash_val is None and show_all:
                continue

            if hash_val is None and hash_function not in ["QUICK", "TLSH"]:
                continue

            value_style = palette.get("hash_value", "magenta")
            display_value = hash_val

            if hash_function == "QUICK" and hash_val is None:
                min_size_mb = QuickHasher.min_permitted_filesize // MiB
                display_value = f"File size below {min_size_mb}MiB minimum"
                value_style = palette.get("info", "dim")
            elif hash_function == "TLSH" and hash_val is None:
                display_value = "The 'py-tlsh' package is not installed"
                value_style = palette.get("info", "dim")

            hash_grid.add_row(f"{hash_function}:", Text(str(display_value), style=value_style))

        console.print(
            Panel(
                hash_grid,
                title=f"[{palette['panel_title']}]🔑 Hashes for {escape(path.name)}[/]",
                expand=False,
                border_style=palette["panel_border"],
            )
        )



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typer
import pytest
from pathlib import Path
from typer.testing import CliRunner
from dorsal.cli.file_app.report_cmd import make_file_report
from dorsal.common.exceptions import DorsalError
from dorsal.common.cli import EXIT_CODE_ERROR


@pytest.fixture
def test_app():
    app = typer.Typer()

    @app.callback()
    def main(ctx: typer.Context):
        ctx.obj = {"palette": {"primary_value": "green", "key": "cyan", "panel_title": "white", "panel_border": "blue"}}

    app.command(name="report")(make_file_report)
    return app


@pytest.fixture
def runner():
    return CliRunner()


def test_report_success_default(runner, test_app, tmp_path, mocker):
    """
    Test happy path: existing file, default output, implicit cache settings.
    """
    target_file = tmp_path / "data.csv"
    target_file.write_text("content")

    mock_generate = mocker.patch("dorsal.api.file.generate_html_file_report")
    mock_console = mocker.patch("dorsal.common.cli.get_rich_console")
    mocker.patch("dorsal.common.cli.determine_use_cache_value", return_value=True)

    result = runner.invoke(test_app, ["report", str(target_file)])

    assert result.exit_code == 0

    mock_generate.assert_called_once()
    call_args = mock_generate.call_args[1]
    assert call_args["file_path"] == str(target_file)
    assert call_args["use_cache"] is True
    assert "scan" in call_args["output_path"]

    # Success messages typically go to stdout
    assert "Report saved successfully" in str(mock_console.return_value.print.call_args)


def test_report_conflict_flags(runner, test_app, tmp_path):
    """
    Test that mutually exclusive flags (--use-cache and --skip-cache) trigger an exit.
    """
    target_file = tmp_path / "test.txt"
    target_file.touch()

    result = runner.invoke(test_app, ["report", str(target_file), "--use-cache", "--skip-cache"])

    assert result.exit_code != 0
    assert "Error: --use-cache and --skip-cache cannot be used together" in result.stderr


def test_report_custom_output_directory(runner, test_app, tmp_path, mocker):
    target_file = tmp_path / "dataset.csv"
    target_file.touch()

    output_dir = tmp_path / "reports"
    output_dir.mkdir()

    mock_generate = mocker.patch("dorsal.api.file.generate_html_file_report")
    mocker.patch("dorsal.common.cli.get_rich_console")

    result = runner.invoke(test_app, ["report", str(target_file), "--output", str(output_dir)])

    assert result.exit_code == 0

    actual_output = mock_generate.call_args[1]["output_path"]
    assert str(output_dir) in actual_output
    assert "dataset_report.html" in actual_output


def test_report_api_error_handling(runner, test_app, tmp_path, mocker):
    """
    Test that exceptions raised by the logic layer are caught and formatted nicely.
    """
    target_file = tmp_path / "corrupt.file"
    target_file.touch()

    mocker.patch("dorsal.api.file.generate_html_file_report", side_effect=DorsalError("File is encrypted"))
    mocker.patch("dorsal.common.cli.get_rich_console")

    result = runner.invoke(test_app, ["report", str(target_file)])

    assert result.exit_code != 0
    assert "Failed to generate report" in result.stderr
    assert "File is encrypted" in result.stderr


def test_report_unexpected_error(runner, test_app, tmp_path, mocker):
    """
    Test handling of generic/unexpected exceptions.
    """
    target_file = tmp_path / "oops.file"
    target_file.touch()

    mocker.patch(
        "dorsal.api.file.generate_html_file_report",
        side_effect=ValueError("Something went really wrong"),
    )
    mocker.patch("dorsal.common.cli.get_rich_console")

    result = runner.invoke(test_app, ["report", str(target_file)])

    assert result.exit_code != 0
    assert "An unexpected error occurred" in result.stderr


def test_report_open_in_browser(runner, test_app, tmp_path, mocker):
    target_file = tmp_path / "viz.data"
    target_file.touch()

    mocker.patch("dorsal.api.file.generate_html_file_report")
    mocker.patch("dorsal.common.cli.get_rich_console")

    mock_browser = mocker.patch("webbrowser.open")

    result = runner.invoke(test_app, ["report", str(target_file), "--open"])

    assert result.exit_code == 0
    mock_browser.assert_called_once()
    assert "file://" in mock_browser.call_args[0][0]


def test_file_not_found_handled_by_typer(runner, test_app):
    """
    Test that Typer automatically handles non-existent files
    because of `exists=True` in the Argument definition.
    """
    result = runner.invoke(test_app, ["report", "ghost_file.txt"])

    assert result.exit_code != 0
    assert "does not exist" in result.stderr



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import logging
import os
from typing import Any, TYPE_CHECKING

from dorsal.common.model import AnnotationModel
from dorsal.file.annotation_models.office_document.config import OFFICE_FORMAT_MAPPING

from dorsal.file.annotation_models.office_document.utils_docx import extract_docx_metadata
from dorsal.file.annotation_models.office_document.utils_xlsx import extract_xlsx_metadata
from dorsal.file.annotation_models.office_document.utils_pptx import extract_pptx_metadata

if TYPE_CHECKING:
    from dorsal.file.configs.model_runner import RunModelResult

logger = logging.getLogger(__name__)

OFFICE_MEDIA_TYPE_MAPPING = {
    "application/vnd.openxmlformats-officedocument.wordprocessingml.document": (
        extract_docx_metadata,
        "docx_stdlib_xml",
    ),
    "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet": (extract_xlsx_metadata, "xlsx_stdlib_xml"),
    "application/vnd.openxmlformats-officedocument.presentationml.presentation": (
        extract_pptx_metadata,
        "pptx_stdlib_xml",
    ),
    None: None,
}


class OfficeDocumentAnnotationModel(AnnotationModel):
    """
    Extracts metadata from Microsoft Office formats (OOXML: .docx, .xlsx, .pptx).
    This model acts as a dispatcher, calling the correct stdlib-based parser.
    """

    id: str = "dorsal/office"
    version: str = "1.0.0"
    variant: str = "dispatcher"

    def main(self) -> dict[str, Any] | None:
        """
        Dispatches to the correct format-specific parser based on media_type.
        """
        logger.debug(
            "OfficeAnnotationModel: Starting metadata extraction for '%s'",
            self.file_path,
        )

        media_type = self.media_type
        parser_info = OFFICE_MEDIA_TYPE_MAPPING.get(media_type)
        metadata: dict[str, Any] | None = None

        if parser_info:
            parser_func, variant_name = parser_info
            self.variant = variant_name
            metadata = parser_func(self.file_path)
        else:
            logger.debug("OfficeAnnotationModel: Skipping. Media type '%s' is not an OOXML office file.", media_type)
            return None

        if metadata is None:
            if self.error is None:
                self.error = f"Failed to parse metadata for file: {self.file_path} (parser: {self.variant})"
            logger.warning(self.error)
            return None

        logger.debug(
            "OfficeAnnotationModel: Successfully processed '%s' with parser '%s'",
            self.file_path,
            self.variant,
        )
        return metadata



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import json
from typer.testing import CliRunner
from unittest.mock import MagicMock
from rich.panel import Panel

from dorsal.cli import app

runner = CliRunner()

MOCK_OPTIMIZE_RESULTS = {
    "stale_records_removed": 15,
    "records_rewritten_for_compression": 100,
    "size_before_bytes": 20480,
    "size_after_bytes": 10240,
    "size_reclaimed_bytes": 10240,
}


@pytest.fixture
def mock_optimize_cache_cmd(mocker):
    """Mocks dependencies for the `cache optimize` command."""
    # Patch at the source due to lazy loading in the command
    mock_get_cache = mocker.patch("dorsal.session.get_shared_cache")

    # Configure the mock cache instance that the getter will return
    mock_cache_instance = MagicMock()
    mock_cache_instance.optimize.return_value = MOCK_OPTIMIZE_RESULTS
    mock_get_cache.return_value = mock_cache_instance

    return {"get_cache": mock_get_cache, "cache_instance": mock_cache_instance}


def test_optimize_cache_panel_output(mock_rich_console, mock_optimize_cache_cmd):
    """Tests the default command, expecting a Rich Panel summary."""
    result = runner.invoke(app, ["cache", "optimize"])

    assert result.exit_code == 0
    mock_optimize_cache_cmd["cache_instance"].optimize.assert_called_once()

    # Verify that a Panel was the object printed to the console
    printed_object = mock_rich_console.print.call_args.args[0]
    assert isinstance(printed_object, Panel)
    assert "Cache Optimization Complete" in str(printed_object.title)


def test_optimize_cache_json_output(mock_rich_console, mock_optimize_cache_cmd):
    """Tests the --json output flag."""
    result = runner.invoke(app, ["cache", "optimize", "--json"])

    assert result.exit_code == 0
    mock_optimize_cache_cmd["cache_instance"].optimize.assert_called_once()
    mock_rich_console.print.assert_called_once()

    # Verify the output is the JSON representation of our mock results
    json_str = mock_rich_console.print.call_args.args[0]
    data = json.loads(json_str)
    assert data == MOCK_OPTIMIZE_RESULTS


def test_optimize_cache_exception_handling(mock_optimize_cache_cmd):
    """Tests that a generic exception is handled gracefully."""
    mock_optimize_cache_cmd["cache_instance"].optimize.side_effect = Exception("General corruption")

    result = runner.invoke(app, ["cache", "optimize"])

    assert result.exit_code != 0
    assert "An error occurred while optimizing the cache: General corruption" in result.output



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import json
import datetime
from typer.testing import CliRunner
from unittest.mock import MagicMock, patch
from rich.panel import Panel
from rich.table import Table


from dorsal.cli import app
from dorsal.cli.dir_app import info_dir_cmd

runner = CliRunner()


TEST_DATA_DIR = "tests/data"

MOCK_STATS_RESULT = {
    "overall": {
        "total_files": 10,
        "total_dirs": 2,
        "hidden_files": 1,
        "total_size": 10485760,  # 10 MiB
        "avg_size": 1048576.0,
        "largest_file": {"size": 5242880, "path": f"{TEST_DATA_DIR}/large.bin"},
        "smallest_file": {"size": 1024, "path": f"{TEST_DATA_DIR}/small.txt"},
        "newest_mod_file": {
            "date": datetime.datetime.now().isoformat(),
            "path": f"{TEST_DATA_DIR}/new.txt",
        },
        "oldest_mod_file": {
            "date": "2023-01-01T12:00:00+00:00",
            "path": f"{TEST_DATA_DIR}/old.txt",
        },
        "oldest_creation_file": {
            "date": "2023-01-01T10:00:00+00:00",
            "path": f"{TEST_DATA_DIR}/archive/first.log",
        },
        "permissions": {"executable": 2, "read_only": 1},
        "time_taken_seconds": 0.5,
        "time_taken_mt": 0.1,
    },
    "by_type": [
        {
            "media_type": "video/mp4",
            "count": 2,
            "total_size": 8388608,
            "percentage": 80.0,
        },
        {
            "media_type": "text/plain",
            "count": 8,
            "total_size": 2097152,
            "percentage": 20.0,
        },
    ],
}


@pytest.fixture
def mock_info_cmd(mocker):
    """Mocks backend dependencies for the `dir info` command."""
    mock_get_info = mocker.patch("dorsal.api.file.get_directory_info", return_value=MOCK_STATS_RESULT)
    mock_save_report = mocker.patch.object(info_dir_cmd, "_save_json_report")

    return {
        "get_directory_info": mock_get_info,
        "save_report": mock_save_report,
    }


@patch("rich.table.Table.grid")
def test_info_dir_success_panel_output(mock_table_grid, mock_rich_console, mock_info_cmd, mocker):
    """Tests the default case by inspecting the data added to the Rich Table."""
    mock_summary_table = MagicMock()
    mock_table_grid.return_value = mock_summary_table

    result = runner.invoke(app, ["dir", "info", TEST_DATA_DIR])

    assert result.exit_code == 0
    mock_info_cmd["get_directory_info"].assert_called_once()
    mock_info_cmd["save_report"].assert_not_called()
    assert mock_rich_console.print.call_count == 2
    assert isinstance(mock_rich_console.print.call_args.args[0], Panel)

    all_rows_text = ""
    for call_args in mock_summary_table.add_row.call_args_list:
        all_rows_text += " ".join(str(arg) for arg in call_args.args)

    assert "Total File Count:" in all_rows_text
    assert "10" in all_rows_text


def test_info_dir_success_with_media_type(mock_rich_console, mock_info_cmd):
    """Tests the --media-type flag, expecting an additional table."""
    result = runner.invoke(app, ["dir", "info", TEST_DATA_DIR, "--media-type"])

    assert result.exit_code == 0
    # The command prints a title, the panel, and the media type table
    assert mock_rich_console.print.call_count == 3

    table_output = mock_rich_console.print.call_args.args[0]
    assert isinstance(table_output, Table)
    assert "Media Type Breakdown" in str(table_output.title)


def test_info_dir_success_json_output(mock_rich_console, mock_info_cmd):
    """Tests the --json flag, expecting raw JSON output."""
    result = runner.invoke(app, ["dir", "info", TEST_DATA_DIR, "--json"])

    assert result.exit_code == 0
    mock_rich_console.print.assert_called_once()

    json_output_str = mock_rich_console.print.call_args.args[0]
    data = json.loads(json_output_str)

    assert data["overall"]["total_files"] == 10
    assert data["by_type"][0]["media_type"] == "video/mp4"
    mock_info_cmd["save_report"].assert_not_called()


def test_info_dir_no_files_found(mock_rich_console, mock_info_cmd):
    """Tests graceful handling when no files are found."""
    empty_stats = MOCK_STATS_RESULT.copy()
    empty_stats["overall"]["total_files"] = 0
    empty_stats["overall"]["total_size"] = 0
    mock_info_cmd["get_directory_info"].return_value = empty_stats

    result = runner.invoke(app, ["dir", "info", TEST_DATA_DIR])

    assert result.exit_code == 0
    mock_rich_console.print.assert_called_once()
    assert "No files found or accessible" in mock_rich_console.print.call_args.args[0]


def test_info_dir_exception_handling(mock_info_cmd):
    """Tests that a generic exception is handled gracefully."""
    mock_info_cmd["get_directory_info"].side_effect = Exception("A critical error occurred")

    result = runner.invoke(app, ["dir", "info", TEST_DATA_DIR])

    assert result.exit_code != 0
    assert "A critical error occurred" in result.output


def test_info_dir_not_a_directory_error(mock_info_cmd):
    """Tests that a NotADirectoryError is handled gracefully."""
    error_msg = f"The specified path is not a directory: {TEST_DATA_DIR}"
    mock_info_cmd["get_directory_info"].side_effect = NotADirectoryError(error_msg)

    result = runner.invoke(app, ["dir", "info", TEST_DATA_DIR])

    assert result.exit_code != 0
    assert error_msg in result.output



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file contains shared fixtures and configuration for the pytest suite, specifically the CLI.

Problem:

`logging.basicConfig()`sets up a logging handler that attaches to the buffer for the *first test*.
In subsequent CLI tests, this handler still exists in memory but its buffer is now closed.
This causes a `ValueError: I/O operation on closed file` when a log message is emitted in a later test.

**Note:** This is a `CliRunner`I/O redirection issue and does not affect the CLI when run normally.

The `clean_logging` fixture (defined below with `autouse=True`) solves this testing issue.
It runs before every test and clears all existing handlers from the root logger.
This forces the CLI to re-initialize logging correctly for each new test's unique I/O buffer.

- To test command logic, arguments, and `print()` output, use the `CliRunner` and the `mock_rich_console` fixture.
- To test messages from `logging` use the `caplog` fixture.

"""

import pytest
import os
import logging
from unittest.mock import MagicMock
from rich.console import Console

from dorsal.common import constants
from dorsal.common import cli as common_cli
from dorsal.session import clear_shared_cache


@pytest.fixture(scope="session", autouse=True)
def global_disable_cache():
    """
    Sets the env var to disable cache for the entire test session.
    This prevents accidental SQLite file creation/locking.
    """
    os.environ[constants.ENV_DORSAL_CACHE_ENABLED] = "false"
    yield


@pytest.fixture(autouse=True)
def reset_dorsal_singletons():
    """
    Ensures every test starts with a clean slate.
    Closes any open cache connections from previous tests.
    """
    clear_shared_cache()
    yield
    clear_shared_cache()


@pytest.fixture(autouse=True)
def clean_logging():
    """
    (Auto-used) Clears all logging handlers before each test.
    """
    root_logger = logging.getLogger()
    if root_logger.hasHandlers():
        root_logger.handlers.clear()
    yield


@pytest.fixture
def mock_rich_console(mocker):
    mock_console = mocker.MagicMock(spec=Console)

    mocker.patch.object(common_cli, "_console_instance", mock_console)
    return mock_console


@pytest.fixture
def mock_auth_app(mocker):
    """
    Mocks all backend dependencies for the `dorsal auth` commands.

    This isolates the CLI layer for focused testing of command logic, argument
    parsing, and user output.
    """
    mocker.patch("dorsal.session.get_shared_dorsal_client")
    mocker.patch("dorsal.common.config.load_config")
    mocker.patch("dorsal.session.clear_shared_dorsal_client")
    mocker.patch("dorsal.common.auth.write_auth_config")
    mocker.patch("dorsal.common.auth.remove_api_key")
    mocker.patch("dorsal.common.auth.get_api_key_from_env", return_value=None)
    mocker.patch("dorsal.client.DorsalClient")



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import subprocess
import logging
from typing import Literal

from dorsal.api.config import unregister_model
from dorsal.common.exceptions import DorsalError
from dorsal.registry.resolution import resolve_target

logger = logging.getLogger(__name__)


def _run_pip_uninstall_streaming(package_name: str) -> bool:
    """
    Runs pip uninstall in a subprocess.
    Returns True if a package was actually removed, False otherwise.
    """
    logger.info(f"Attempting to uninstall python package: {package_name}...")

    cmd = [sys.executable, "-m", "pip", "uninstall", "-y", package_name]

    is_verbose = logger.isEnabledFor(logging.INFO)

    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
        encoding="utf-8",
        errors="replace",
        bufsize=1,
    )

    captured_lines = []
    skipped = False

    if process.stdout:
        for line in process.stdout:
            if f"Skipping {package_name} as it is not installed" in line:
                skipped = True

            if is_verbose:
                sys.stdout.write(line)
            captured_lines.append(line)

    return_code = process.wait()

    if return_code != 0:
        full_log = "".join(captured_lines)
        if not is_verbose:
            sys.stderr.write(f"\n[!] Pip Uninstall Log for '{package_name}':\n")
            sys.stderr.write(full_log)
        logger.warning(f"Pip reported an error uninstalling '{package_name}'.")
        return False

    return not skipped


def uninstall_model_target(
    target: str,
    scope: Literal["project", "global"] = "project",
) -> str:
    """
    Uninstalls a model.
    Raises DorsalError if NO action was taken (nothing removed from config AND nothing removed from pip).
    """
    logger.info(f"Processing uninstall target: {target}")

    _, safe_package_name = resolve_target(target)

    did_modify_config = False
    did_uninstall_pip = False

    try:
        unregister_model(package_name=safe_package_name, scope=scope)
        did_modify_config = True
    except KeyError:
        logger.debug(f"Config cleanup skipped: '{safe_package_name}' not found in {scope} config.")
    except Exception as e:
        raise DorsalError(f"Failed to update dorsal configuration: {e}") from e

    did_uninstall_pip = _run_pip_uninstall_streaming(safe_package_name)

    if not did_modify_config and not did_uninstall_pip:
        raise DorsalError(
            f"Could not find model '{target}' (resolved as package: '{safe_package_name}') "
            f"in the {scope} configuration OR installed in the environment."
        )

    action_summary = []
    if did_modify_config:
        action_summary.append(f"removed from {scope} config")
    if did_uninstall_pip:
        action_summary.append("uninstalled python package")

    logger.info(f"Successfully uninstalled {safe_package_name} ({', '.join(action_summary)})")
    return safe_package_name



# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

PYPROJECT_TEMPLATE = """[build-system]
requires = ["hatchling"]
build-backend = "hatchling.build"

[project]
name = "{package_name}"
version = "0.1.0"
description = "Dorsal model: {human_name}"
readme = "README.md"
requires-python = ">=3.11"
dependencies = [
    "dorsalhub",
]

[dependency-groups]
dev = [
    "pytest",
    "pytest-cov",
    "ruff",
    "mypy"
]

[project.entry-points."dorsal.models"]
{entry_point_name} = "{module_name}"

[tool.hatch.build.targets.wheel]
packages = ["{module_name}"]

[tool.hatch.build.targets.wheel.force-include]
"model_config.toml" = "{module_name}/model_config.toml"
"""

MODEL_CONFIG_TEMPLATE = """# Dorsal Model Configuration
# Documentation: https://docs.dorsalhub.com/models/config

# --- BASIC CONFIG ---
# The name of your Annotation Model class
model_class = "{class_name}"

# The schema which validates the model output
schema_id = "open/generic"
schema_version = "{schema_version}"

# --- DEPENDENCIES ---
# Define when this model should run.

# Example 1 (media_type): Run on files with the `text/plain` media type
[[dependencies]]
type = "media_type"
include = ["text/plain"]

# Example 2 (media_type): Run on any audio/video files except for Matroska video
# [[dependencies]]
# type = "media_type"
# include = ["audio", "video", "text/plain"]
# exclude = ["video/matroska"]

# Example 3 (file extension): Run only on MS Office Docs
# [[dependencies]]
# type = "extension"
# extensions = ["doc", "docx]

# Example 4 (file size): Run only on files smaller than 100MB
# [[dependencies]]
# type = "file_size"
# max_size = "100MB"

# --- RUNTIME OPTIONS ---
[options]
# Pass keyword arguments to your model class `main` method
"""

INIT_TEMPLATE = """from .model import {class_name}

__all__ = ["{class_name}"]
"""

MODEL_TEMPLATE = """from importlib.metadata import version
from dorsal import AnnotationModel

class {class_name}(AnnotationModel):
    id = "local/{entry_point_name}"
    version = version("{package_name}")

    def main(self):
        
        # Implement your logic here.

        return {{"data": {{"Hello": "World"}}}}
"""

TEST_TEMPLATE = """import pytest
import pathlib
import tomllib
from dorsal.testing import run_model
from {module_name}.model import {class_name}

def test_model_integration(tmp_path):
    # 1. Setup a dummy file
    dummy_file = tmp_path / "test_doc.txt"
    dummy_file.write_text("Hello Dorsal!")

    # 2. Load Registry Metadata
    root = pathlib.Path(__file__).parent.parent
    with open(root / "{module_name}" / "model_config.toml", "rb") as f:
        config = tomllib.load(f)

    # 3. Run Model
    result = run_model(
        annotation_model={class_name},
        file_path=str(dummy_file),
        schema_id=config.get("schema_id", "open/generic"),
        schema_version=config.get("schema_version"),
        dependencies=config.get("dependencies"),
        options=config.get("options"),
    )

    assert result.error is None, f"Model execution failed: {{result.error}}"
    assert result.record is not None
"""



# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from typing import Type, Any, Union
import re

from pydantic import BaseModel, Field, HttpUrl, field_validator

from dorsal.common.model import AnnotationModel
from dorsal.file.configs.model_runner import DependencyType

REGISTRY_ID_RX = re.compile(r"^[a-z0-9][a-z0-9._-]*\/[a-z0-9][a-z0-9._-]*$", re.IGNORECASE)
SAFE_INSTALL_URL_RX = re.compile(r"^git\+https?://[a-zA-Z0-9.\-_/]+\.git@[a-f0-9]{40}$")


class ModelSpec(BaseModel):
    """Specification for external (installed) model packages."""

    model_class: Type[AnnotationModel]
    schema_id: str
    schema_version: str | None = None
    dependencies: list[DependencyType] | None = None
    validation_model: Any | None = None
    options: dict[str, Any] | None = None

    @field_validator("model_class")
    @classmethod
    def validate_model_class(cls, v):
        if not issubclass(v, AnnotationModel):
            raise ValueError(f"External model_class must inherit from dorsal.AnnotationModel, got {v}")
        return v


class InitResult(BaseModel):
    path: pathlib.Path
    package_name: str
    clean_name: str


def is_registry_id(target: str) -> bool:
    """Determines if a target string is a valid Registry ID."""
    if "/" not in target:
        return False

    return bool(REGISTRY_ID_RX.match(target))


def is_valid_local_path(target: str) -> bool:
    """
    Determines if the target is a valid local resource (File or Directory).
    """
    path = pathlib.Path(target)
    return path.exists() and (path.is_dir() or path.is_file())


def validate_install_url(url: str) -> str:
    """
    Validates the install URL returned by the Registry API.
        1. Protocol must be git+http(s).
        2. Must end in .git@{sha} (pinned version).
    """
    if not SAFE_INSTALL_URL_RX.match(url):
        raise ValueError(
            f"Registry returned an invalid install URL: '{url}'. "
            "Expected format: 'git+https://host/repo.git@commit_sha'"
        )
    return url
